"""Collection of tests for :mod:`orion.plotting.backend_plotly`."""
import copy

import numpy
import pandas
import plotly
import pytest

import orion.client
from orion.analysis.partial_dependency_utils import partial_dependency_grid
from orion.core.worker.experiment import Experiment
from orion.plotting.base import (
    lpi,
    parallel_coordinates,
    partial_dependencies,
    rankings,
    regret,
    regrets,
)
from orion.testing import create_experiment
from orion.testing.plotting import (
    assert_lpi_plot,
    assert_parallel_coordinates_plot,
    assert_partial_dependencies_plot,
    assert_rankings_plot,
    assert_regret_plot,
    assert_regrets_plot,
)

config = dict(
    name="experiment-name",
    space={"x": "uniform(0, 200)"},
    metadata={
        "user": "test-user",
        "orion_version": "XYZ",
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir="",
    algorithms={"random": {"seed": 1}},
    producer={"strategy": "NoParallelStrategy"},
)

trial_config = {
    "experiment": 0,
    "status": "completed",
    "worker": None,
    "start_time": None,
    "end_time": None,
    "heartbeat": None,
    "results": [],
    "params": [],
}


def mock_space(x="uniform(0, 6)", y="uniform(0, 3)", **kwargs):
    """Build a mocked space"""
    mocked_config = copy.deepcopy(config)
    mocked_config["space"] = {"x": x}
    if y is not None:
        mocked_config["space"]["y"] = y
    mocked_config["space"].update(kwargs)
    return mocked_config


def mock_experiment(
    monkeypatch, ids=None, x=None, y=None, z=None, objectives=None, status=None
):
    """Mock experiment to_pandas to return given data (or default one)"""
    if ids is None:
        ids = ["a", "b", "c", "d"]
    if x is None:
        x = [0, 1, 2, 4]
    if y is None:
        y = [3, 2, 0, 1]
    if objectives is None:
        objectives = [0.1, 0.2, 0.3, 0.5]
    if status is None:
        status = ["completed", "completed", "completed", "completed"]

    data = {
        "id": ids,
        "x": x,
        "objective": objectives,
        "status": status,
        "suggested": ids,
    }

    if not isinstance(y, str):
        data["y"] = y
    if z is not None:
        data["z"] = z

    def to_pandas(self, with_evc_tree=False):

        return pandas.DataFrame(data=data)

    monkeypatch.setattr(Experiment, "to_pandas", to_pandas)


def mock_experiment_with_random_to_pandas(monkeypatch, status=None, unbalanced=False):
    def to_pandas(self, with_evc_tree=False):
        if unbalanced:
            N = numpy.random.randint(5, 15)
        elif status is not None:
            N = len(status)
        else:
            N = 10
        ids = numpy.arange(N)
        x = numpy.random.normal(0, 0.1, size=N)
        y = numpy.random.normal(0, 0.1, size=N)
        objectives = numpy.random.normal(0, 0.1, size=N)
        if status is None:
            exp_status = ["completed"] * N
        else:
            exp_status = status

        data = pandas.DataFrame(
            data={
                "id": ids,
                "x": x,
                "y": y,
                "objective": objectives,
                "status": exp_status,
                "suggested": ids,
            }
        )

        return data

    monkeypatch.setattr(Experiment, "to_pandas", to_pandas)


def mock_model():
    """Return a mocked regressor which just predict iterated integers"""

    class Model:
        """Mocked Regressor"""

        def __init__(self):
            self.i = 0

        def predict(self, data):
            """Returns counting of predictions requested."""
            data = numpy.arange(data.shape[0]) + self.i
            self.i += data.shape[0]
            return data  #  + numpy.random.normal(0, self.i, size=data.shape[0])

    return Model()


def mock_train_regressor(monkeypatch, assert_model=None, assert_model_kwargs=None):
    """Mock the train_regressor to return the mocked regressor instead"""

    def train_regressor(model, data, **kwargs):
        """Return the mocked model, and then model argument if requested"""
        if assert_model:
            assert model == assert_model
        if assert_model_kwargs:
            assert kwargs == assert_model_kwargs
        return mock_model()

    monkeypatch.setattr(
        "orion.analysis.partial_dependency_utils.train_regressor", train_regressor
    )


@pytest.mark.usefixtures("version_XYZ")
class TestLPI:
    """Tests the ``lpi()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            lpi(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self, monkeypatch):
        """Tests the layout of the plot"""
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))
            df = experiment.to_pandas()
            assert df["x"].tolist() == [0, 1, 2, 4]
            assert df["y"].tolist() == [3, 2, 0, 1]
            assert df["objective"].tolist() == [0.1, 0.2, 0.3, 0.5]

        assert_lpi_plot(plot, dims=["x", "y"])

    def test_experiment_worker_as_parameter(self, monkeypatch):
        """Tests that ``Experiment`` is a valid parameter"""
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            experiment,
            _,
        ):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=["x", "y"])

    def test_ignore_uncompleted_statuses(self, monkeypatch):
        """Tests that uncompleted statuses are ignored"""
        config = mock_space()
        mock_experiment(
            monkeypatch,
            ids="abcdefgh",
            x=[0, 0, 0, 1, 0, 2, 0, 3],
            y=[1, 0, 0, 2, 0, 0, 0, 3],
            objectives=[0.1, None, None, 0.2, None, 0.3, None, 0.5],
            status=[
                "completed",
                "new",
                "reserved",
                "completed",
                "broken",
                "completed",
                "interrupted",
                "completed",
            ],
        )
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment)

        assert_lpi_plot(plot, dims=["x", "y"])

    def test_multidim(self, monkeypatch):
        """Tests that dimensions with shape > 1 are flattened properly"""
        config = mock_space(y="uniform(0, 3, shape=2)")
        mock_experiment(monkeypatch, y=[[3, 3], [2, 3], [1, 2], [0, 3]])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=["x", "y[0]", "y[1]"])

    def test_fidelity(self, monkeypatch):
        """Tests that fidelity is supported"""
        config = mock_space(y="fidelity(1, 200, base=3)")
        mock_experiment(monkeypatch, y=[1, 3 ** 2, 1, 3 ** 4])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=["x", "y"])

    def test_categorical(self, monkeypatch):
        """Tests that categorical is supported"""
        config = mock_space(y='choices(["a", "b", "c"])')
        mock_experiment(monkeypatch, y=["c", "c", "a", "b"])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=["x", "y"])

    def test_categorical_multidim(self, monkeypatch):
        """Tests that multidim categorical is supported"""
        config = mock_space(y='choices(["a", "b", "c"], shape=3)')
        mock_experiment(
            monkeypatch,
            y=[["c", "b", "a"], ["c", "a", "c"], ["a", "b", "a"], ["c", "b", "b"]],
        )

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=["x", "y[0]", "y[1]", "y[2]"])


@pytest.mark.usefixtures("version_XYZ")
class TestParallelCoordinates:
    """Tests the ``parallel_coordinates()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            parallel_coordinates(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = parallel_coordinates(experiment)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self):
        """Tests the layout of the plot"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=["x", "loss"])

    def test_experiment_worker_as_parameter(self):
        """Tests that ``Experiment`` is a valid parameter"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            experiment,
            _,
        ):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=["x", "loss"])

    def test_ignore_uncompleted_statuses(self):
        """Tests that uncompleted statuses are ignored"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=["x", "loss"])

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                parallel_coordinates(experiment, order=["unsupported"])

    def test_order_columns(self):
        """Tests that columns are sorted according to ``order``"""
        multidim_config = copy.deepcopy(config)
        for k in "yzutv":
            multidim_config["space"][k] = "uniform(0, 200)"
        with create_experiment(multidim_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment, order="vzyx")

        assert_parallel_coordinates_plot(plot, order=["v", "z", "y", "x", "loss"])

    def test_multidim(self):
        """Tests that dimensions with shape > 1 are flattened properly"""
        multidim_config = copy.deepcopy(config)
        multidim_config["space"]["y"] = "uniform(0, 200, shape=4)"
        with create_experiment(multidim_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(
            plot, order=["x", "y[0]", "y[1]", "y[2]", "y[3]", "loss"]
        )

    def test_fidelity(self):
        """Tests that fidelity is set to first column by default"""
        fidelity_config = copy.deepcopy(config)
        fidelity_config["space"]["z"] = "fidelity(1, 200, base=3)"
        with create_experiment(fidelity_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=["z", "x", "loss"])

    def test_categorical(self):
        """Tests that categorical is supported"""
        categorical_config = copy.deepcopy(config)
        categorical_config["space"]["z"] = 'choices(["a", "b", "c"])'
        with create_experiment(categorical_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=["x", "z", "loss"])

    def test_categorical_multidim(self):
        """Tests that multidim categorical is supported"""
        categorical_config = copy.deepcopy(config)
        categorical_config["space"]["z"] = 'choices(["a", "b", "c"], shape=3)'
        with create_experiment(categorical_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(
            plot, order=["x", "z[0]", "z[1]", "z[2]", "loss"]
        )


@pytest.mark.usefixtures("version_XYZ")
class TestPartialDependencies:
    """Tests the ``partial_dependencies()`` method provided by the plotly backend"""

    def test_returns_plotly_object(self, monkeypatch):
        """Tests that the plotly backend returns a plotly object"""
        mock_train_regressor(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self, monkeypatch):
        """Tests the layout of the plot"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )
            df = experiment.to_pandas()
            assert df["x"].tolist() == [0, 1, 2, 4]
            assert df["y"].tolist() == [3, 2, 0, 1]
            assert df["objective"].tolist() == [0.1, 0.2, 0.3, 0.5]

        assert_partial_dependencies_plot(plot, dims=["x", "y"])

    def test_ignore_uncompleted_statuses(self, monkeypatch):
        """Tests that uncompleted statuses are ignored"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(
            monkeypatch,
            ids="abcdefgh",
            x=[0, 0, 0, 1, 0, 2, 0, 3],
            y=[1, 0, 0, 2, 0, 0, 0, 3],
            objectives=[0.1, None, None, 0.2, None, 0.3, None, 0.5],
            status=[
                "completed",
                "new",
                "reserved",
                "completed",
                "broken",
                "completed",
                "interrupted",
                "completed",
            ],
        )
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(experiment, n_grid_points=5)

        assert_partial_dependencies_plot(plot, dims=["x", "y"])

    def test_multidim(self, monkeypatch):
        """Tests that dimensions with shape > 1 are flattened properly"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y="uniform(0, 3, shape=2)")
        mock_experiment(monkeypatch, y=[[3, 3], [2, 3], [1, 2], [0, 3]])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(plot, dims=["x", "y[0]", "y[1]"])

    def test_fidelity(self, monkeypatch):
        """Tests that fidelity is supported"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y="fidelity(1, 200, base=3)")
        mock_experiment(monkeypatch, y=[1, 3 ** 2, 1, 3 ** 4])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(plot, dims=["x", "y"], log_dims=["y"])

    def test_categorical(self, monkeypatch):
        """Tests that categorical is supported"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y='choices(["a", "b", "c"])')
        mock_experiment(monkeypatch, y=["c", "c", "a", "b"])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        # There is only 3 categories, so test must be adjusted accordingly.
        assert_partial_dependencies_plot(
            plot, dims=["x", "y"], n_grid_points={"x": 5, "y": 3}
        )

    def test_categorical_multidim(self, monkeypatch):
        """Tests that multidim categorical is supported"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y='choices(["a", "b", "c"], shape=3)')
        mock_experiment(
            monkeypatch,
            y=[["c", "b", "a"], ["c", "a", "c"], ["a", "b", "a"], ["c", "b", "b"]],
        )

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(
            plot,
            dims=["x", "y[0]", "y[1]", "y[2]"],
            n_grid_points={"x": 5, "y[0]": 3, "y[1]": 3, "y[2]": 3},
        )

    def test_logarithmic_scales_first(self, monkeypatch):
        """Test that log dims are turn to log scale

        Test first dim specifically because special xaxis name for first dim.
        """
        mock_train_regressor(monkeypatch)
        config = mock_space(x="loguniform(0.001, 1)", z="uniform(0, 1)")
        mock_experiment(monkeypatch, x=[0.001, 0.1, 0.01, 1], z=[0, 0.1, 0.2, 0.5])

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(
            plot, dims=["x", "y", "z"], n_grid_points=5, log_dims=["x"]
        )

    def test_logarithmic_scales_any_dim(self, monkeypatch):
        """Test that log dims are turn to log scale"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y="loguniform(0.001, 1)", z="uniform(0, 1)")
        mock_experiment(monkeypatch, y=[0.001, 0.1, 0.01, 1], z=[0, 0.1, 0.2, 0.5])

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(
            plot, dims=["x", "y", "z"], n_grid_points=5, log_dims=["y"]
        )

    def test_int_logarithmic_scales(self, monkeypatch):
        """Test that int log dims are turn to log scale"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y="loguniform(1, 1000, discrete=True)", z="uniform(0, 1)")
        mock_experiment(monkeypatch, y=[1, 10, 100, 1000], z=[0, 0.1, 0.2, 0.5])

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(
            plot, dims=["x", "y", "z"], n_grid_points=5, log_dims=["y"]
        )

    def test_one_param(self, monkeypatch):
        """Test ploting a space with only 1 dim"""
        mock_train_regressor(monkeypatch)
        config = mock_space(y=None)
        mock_experiment(monkeypatch, y="drop")

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment, n_grid_points=5, model_kwargs=dict(random_state=1)
            )

        assert_partial_dependencies_plot(plot, dims=["x"], n_grid_points=5)

    def test_select_params(self, monkeypatch):
        """Test selecting subset"""
        mock_train_regressor(monkeypatch)
        config = mock_space(z="uniform(0, 1)")
        mock_experiment(monkeypatch, z=[0, 0.1, 0.2, 0.5])

        for params in [["x"], ["x", "y"], ["y", "z"]]:
            with create_experiment(config, trial_config) as (_, _, experiment):
                plot = partial_dependencies(
                    experiment,
                    params=params,
                    n_grid_points=5,
                    model_kwargs=dict(random_state=1),
                )

            assert_partial_dependencies_plot(plot, dims=params, n_grid_points=5)

    def test_custom_smoothing(self, monkeypatch):
        """Test changing smoothing value"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(monkeypatch)

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=5,
                model_kwargs=dict(random_state=1),
                smoothing=1.2,
            )

        with pytest.raises(AssertionError):
            assert_partial_dependencies_plot(plot, dims=["x", "y"], n_grid_points=5)

        assert_partial_dependencies_plot(
            plot, dims=["x", "y"], n_grid_points=5, smoothing=1.2
        )

    def test_custom_n_grid_points(self, monkeypatch):
        """Test changing n_grid_points value"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(monkeypatch)

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=10,
                model_kwargs=dict(random_state=1),
            )

        with pytest.raises(AssertionError):
            assert_partial_dependencies_plot(plot, dims=["x", "y"], n_grid_points=5)

        assert_partial_dependencies_plot(plot, dims=["x", "y"], n_grid_points=10)

    def test_custom_n_samples(self, monkeypatch):
        """Test changing n_samples value"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(monkeypatch)

        PARAMS = ["x", "y"]
        N_SAMPLES = numpy.random.randint(20, 50)

        def mock_partial_dependency_grid(space, model, params, samples, n_points):
            print(samples)
            assert samples.shape == (N_SAMPLES, len(PARAMS))
            return partial_dependency_grid(space, model, params, samples, n_points)

        monkeypatch.setattr(
            "orion.analysis.partial_dependency_utils.partial_dependency_grid",
            mock_partial_dependency_grid,
        )

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=10,
                model_kwargs=dict(random_state=1),
                n_samples=N_SAMPLES,
            )

    def test_custom_colorscale(self, monkeypatch):
        """Test changing colorscale"""
        mock_train_regressor(monkeypatch)
        config = mock_space()
        mock_experiment(monkeypatch)

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=5,
                colorscale="Viridis",
                model_kwargs=dict(random_state=1),
            )

        with pytest.raises(AssertionError):
            assert_partial_dependencies_plot(
                plot, dims=["x", "y"], n_grid_points=5, custom_colorscale=False
            )

        assert_partial_dependencies_plot(
            plot, dims=["x", "y"], n_grid_points=5, custom_colorscale=True
        )

    def test_custom_model(self, monkeypatch):
        """Test changing type of regression model"""
        mock_train_regressor(monkeypatch, assert_model="BaggingRegressor")
        config = mock_space()
        mock_experiment(monkeypatch)

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=5,
                model="BaggingRegressor",
                model_kwargs=dict(random_state=1),
            )

    def test_custom_model_kwargs(self, monkeypatch):
        """Test changing arguments of regression model"""
        mock_train_regressor(monkeypatch, assert_model_kwargs=dict(random_state=1))
        config = mock_space()
        mock_experiment(monkeypatch)

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = partial_dependencies(
                experiment,
                n_grid_points=5,
                model_kwargs=dict(random_state=1),
            )


@pytest.mark.usefixtures("version_XYZ")
class TestRankings:
    """Tests the ``rankings()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            rankings(None)

    def test_returns_plotly_object(self, monkeypatch):
        """Tests that the plotly backend returns a plotly object"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings([experiment, experiment])

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self, monkeypatch):
        """Tests the layout of the plot"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings([experiment])

        assert_rankings_plot(plot, [f"{experiment.name}-v{experiment.version}"])

    def test_list_of_experiments(self, monkeypatch):
        """Tests the rankings with list of experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            child = orion.client.create_experiment(
                experiment.name, branching={"branch_to": "child"}
            )

            plot = rankings([experiment, child])

        # Exps are sorted alphabetically by names.
        assert_rankings_plot(
            plot, [f"{exp.name}-v{exp.version}" for exp in [child, experiment]]
        )

    def test_list_of_experiments_name_conflict(self, monkeypatch):
        """Tests the rankings with list of experiments with the same name"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            child = orion.client.create_experiment(
                experiment.name, branching={"branch_to": experiment.name}
            )
            assert child.name == experiment.name
            assert child.version == experiment.version + 1
            plot = rankings([experiment, child])

        # Exps are sorted alphabetically by names.
        assert_rankings_plot(
            plot, [f"{exp.name}-v{exp.version}" for exp in [experiment, child]]
        )

    def test_dict_of_experiments(self, monkeypatch):
        """Tests the rankings with renamed experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings({"exp-1": experiment, "exp-2": experiment})

        assert_rankings_plot(plot, ["exp-1", "exp-2"])

    def test_list_of_dict_of_experiments(self, monkeypatch):
        """Tests the rankings with avg of competitions"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings(
                [{"exp-1": experiment, "exp-2": experiment} for _ in range(10)]
            )

        assert_rankings_plot(plot, ["exp-1", "exp-2"], with_avg=True)

    def test_dict_of_list_of_experiments(self, monkeypatch):
        """Tests the rankings with avg of experiments separated in lists"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings({"exp-1": [experiment] * 10, "exp-2": [experiment] * 10})

        assert_rankings_plot(plot, ["exp-1", "exp-2"], with_avg=True)

    def test_unbalanced_experiments(self, monkeypatch):
        """Tests the regrets with avg of unbalanced experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch, unbalanced=True)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = rankings({"exp-1": [experiment] * 10, "exp-2": [experiment] * 10})

        assert_rankings_plot(plot, ["exp-1", "exp-2"], with_avg=True, balanced=0)

    def test_ignore_uncompleted_statuses(self, monkeypatch):
        """Tests that uncompleted statuses are ignored"""
        mock_experiment_with_random_to_pandas(
            monkeypatch,
            status=[
                "completed",
                "new",
                "reserved",
                "completed",
                "broken",
                "completed",
                "interrupted",
                "completed",
            ],
        )
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = rankings([experiment])

        assert_rankings_plot(
            plot, [f"{experiment.name}-v{experiment.version}"], balanced=4
        )

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                rankings([experiment], order_by="unsupported")


@pytest.mark.usefixtures("version_XYZ")
class TestRegret:
    """Tests the ``regret()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            regret(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regret(experiment)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self):
        """Tests the layout of the plot"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_experiment_worker_as_parameter(self):
        """Tests that ``Experiment`` is a valid parameter"""
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            experiment,
            _,
        ):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_ignore_uncompleted_statuses(self):
        """Tests that uncompleted statuses are ignored"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                regret(experiment, order_by="unsupported")


@pytest.mark.usefixtures("version_XYZ")
class TestRegrets:
    """Tests the ``regrets()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            regrets(None)

    def test_returns_plotly_object(self, monkeypatch):
        """Tests that the plotly backend returns a plotly object"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regrets([experiment])

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self, monkeypatch):
        """Tests the layout of the plot"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regrets([experiment])

        assert_regrets_plot(plot, [f"{experiment.name}-v{experiment.version}"])

    def test_list_of_experiments(self, monkeypatch):
        """Tests the regrets with list of experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            child = orion.client.create_experiment(
                experiment.name, branching={"branch_to": "child"}
            )

            plot = regrets([experiment, child])

        # Exps are sorted alphabetically by names.
        assert_regrets_plot(
            plot, [f"{exp.name}-v{exp.version}" for exp in [child, experiment]]
        )

    def test_list_of_experiments_name_conflict(self, monkeypatch):
        """Tests the regrets with list of experiments with the same name"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            child = orion.client.create_experiment(
                experiment.name, branching={"branch_to": experiment.name}
            )
            assert child.name == experiment.name
            assert child.version == experiment.version + 1
            plot = regrets([experiment, child])

        # Exps are sorted alphabetically by names.
        assert_regrets_plot(
            plot, [f"{exp.name}-v{exp.version}" for exp in [experiment, child]]
        )

    def test_dict_of_experiments(self, monkeypatch):
        """Tests the regrets with renamed experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regrets({"exp-1": experiment, "exp-2": experiment})

        assert_regrets_plot(plot, ["exp-1", "exp-2"])

    def test_dict_of_list_of_experiments(self, monkeypatch):
        """Tests the regrets with avg of experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regrets({"exp-1": [experiment] * 10, "exp-2": [experiment] * 10})

        assert_regrets_plot(plot, ["exp-1", "exp-2"], with_avg=True)

    def test_unbalanced_experiments(self, monkeypatch):
        """Tests the regrets with avg of unbalanced experiments"""
        mock_experiment_with_random_to_pandas(monkeypatch, unbalanced=True)
        with create_experiment(config, trial_config, ["completed"]) as (
            _,
            _,
            experiment,
        ):
            plot = regrets({"exp-1": [experiment] * 10, "exp-2": [experiment] * 10})

        assert_regrets_plot(plot, ["exp-1", "exp-2"], with_avg=True, balanced=0)

    def test_ignore_uncompleted_statuses(self, monkeypatch):
        """Tests that uncompleted statuses are ignored"""
        mock_experiment_with_random_to_pandas(
            monkeypatch,
            status=[
                "completed",
                "new",
                "reserved",
                "completed",
                "broken",
                "completed",
                "interrupted",
                "completed",
            ],
        )
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = regrets([experiment])

        assert_regrets_plot(
            plot, [f"{experiment.name}-v{experiment.version}"], balanced=4
        )

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                regrets([experiment], order_by="unsupported")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.pickleddb`."""
import os
import pickle
from datetime import datetime

import pytest

from orion.core.io.database import Database, DatabaseError, DuplicateKeyError
from orion.core.io.database.ephemeraldb import EphemeralDB
from orion.core.io.database.mongodb import MongoDB
from orion.core.io.database.pickleddb import PickledDB

_DB_TYPES = ["ephemeraldb", "mongodb", "pickleddb"]


def get_db(orion_db):
    """Get database instance"""
    if isinstance(orion_db, EphemeralDB):
        return orion_db._db
    elif isinstance(orion_db, MongoDB):
        return orion_db._db
    elif isinstance(orion_db, PickledDB):
        return orion_db._get_database()._db
    else:
        raise TypeError("Invalid database type")


def dump_db(orion_db, db):
    """Dump database if necessary"""
    if isinstance(orion_db, EphemeralDB):
        pass
    elif isinstance(orion_db, MongoDB):
        pass
    elif isinstance(orion_db, PickledDB):
        ephemeral_db = orion_db._get_database()
        ephemeral_db._db = db
        orion_db._dump_database(ephemeral_db)
    else:
        raise TypeError("Invalid database type")


@pytest.fixture(scope="module", autouse=True, params=_DB_TYPES)
def db_type(pytestconfig, request):
    """Return the string identifier of a supported database type based on the
    --mongodb option

    If `--mongodb` is active, only MongoDB tests will be run. Otherwise,
    all non-MongoDB will be run.
    """
    if request.param == "mongodb" and not pytestconfig.getoption("--mongodb"):
        pytest.skip("{} tests disabled".format(request.param))
    elif request.param != "mongodb" and pytestconfig.getoption("--mongodb"):
        pytest.skip("{} tests disabled".format(request.param))
    yield request.param


@pytest.fixture(scope="module")
def orion_db(db_type):
    """Return a supported database wrapper instance initiated with test opts."""
    if db_type == "ephemeraldb":
        EphemeralDB.instance = None
        orion_db = EphemeralDB()
    elif db_type == "mongodb":
        MongoDB.instance = None
        orion_db = MongoDB(username="user", password="pass", name="orion_test")
    elif db_type == "pickleddb":
        PickledDB.instance = None
        orion_db = PickledDB(host="orion_db.pkl")
    else:
        raise ValueError("Invalid database type")

    yield orion_db

    if db_type == "ephemeraldb":
        pass
    elif db_type == "mongodb":
        orion_db.close_connection()
    elif db_type == "pickleddb":
        pass


@pytest.fixture()
def clean_db(orion_db):
    """Cleaned the current database prior a test."""
    if isinstance(orion_db, EphemeralDB):
        pass
    elif isinstance(orion_db, MongoDB):
        pass
    elif isinstance(orion_db, PickledDB):
        if os.path.exists(orion_db.host):
            os.remove(orion_db.host)

    database = get_db(orion_db)

    # Drop experiment data
    database["experiments"].drop()
    database["lying_trials"].drop()
    database["trials"].drop()
    database["workers"].drop()
    database["resources"].drop()

    dump_db(orion_db, database)

    yield database


@pytest.fixture()
def db_test_data(request, db_type):
    """Return test data corresponding to the current database type"""
    for key in request.param:
        if db_type in key:
            yield request.param[key]
            break
    else:
        raise ValueError("Invalid database type")


@pytest.fixture(autouse=True)
def insert_collections(request, orion_db, clean_db):
    """Drop a collection prior a test"""
    collections_data = (
        request.node.get_closest_marker("insert_collections").args[0]
        if request.node.get_closest_marker("insert_collections")
        else {}
    )
    for name, data in collections_data.items():
        clean_db[name].drop()
        clean_db[name].insert_many(data)

    dump_db(orion_db, clean_db)

    yield collections_data

    for name in collections_data.keys():
        clean_db[name].drop()

    dump_db(orion_db, clean_db)


@pytest.fixture(autouse=True)
def drop_collections(request, orion_db):
    """Drop a collection prior a test"""
    db = get_db(orion_db)
    collections = (
        request.node.get_closest_marker("drop_collections").args[0]
        if request.node.get_closest_marker("drop_collections")
        else tuple()
    )
    for collection in collections:
        db[collection].drop()
    yield
    for collection in collections:
        db[collection].drop()


@pytest.fixture(autouse=True)
def skip_if_not_db_type(request, db_type):
    """Skip test if th database type does no match the database type marker"""
    db_types_only = request.node.get_closest_marker("db_types_only")
    if db_types_only and db_type not in db_types_only.args[0]:
        pytest.skip("{} test only".format(db_types_only.args[0]))


insert_test_collection = pytest.mark.insert_collections(
    {
        "test_collection": [
            {
                "_id": 0,
                "field0": "same0",
                "field1": "same1",
                "datetime": datetime(2017, 11, 22, 0, 0, 0),
                "same_field": "same",
                "unique_field": "unique0",
                "same_comp": {"ound": "same_compound"},
                "unique_comp": {"ound": "compound0"},
            },
            {
                "_id": 1,
                "field0": "same0",
                "field1": "diff1",
                "datetime": datetime(2017, 11, 23, 0, 0, 0),
                "same_field": "same",
                "unique_field": "unique1",
                "same_comp": {"ound": "same_compound"},
                "unique_comp": {"ound": "compound1"},
            },
            {
                "_id": 2,
                "field0": "diff0",
                "field1": "same1",
                "datetime": datetime(2017, 11, 24, 0, 0, 0),
                "same_field": "same",
                "unique_field": "unique2",
                "same_comp": {"ound": "same_compound"},
                "unique_comp": {"ound": "compound2"},
            },
        ]
    }
)


@pytest.fixture()
def test_collection(insert_collections):
    """Drop a collection prior a test"""
    yield insert_collections["test_collection"]


# TESTS SET


@pytest.mark.usefixtures("clean_db")
@pytest.mark.drop_collections(["new_collection"])
class TestEnsureIndex(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.ensure_index`."""

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": (False, "new_field", "new_field_1", False),
                "mongodb": (False, "new_field", "new_field_1", True),
            },
            {
                "ephemeraldb_pickleddb": (True, "new_field", "new_field_1", True),
                "mongodb": (True, "new_field", "new_field_1", True),
            },
        ],
        indirect=True,
    )
    def test_new_index(self, orion_db, db_test_data):
        """Index should be added to database"""
        unique, key, stored_key, key_present = db_test_data
        assert stored_key not in get_db(orion_db)["new_collection"].index_information()

        orion_db.ensure_index("new_collection", key, unique=unique)
        assert (
            stored_key in get_db(orion_db)["new_collection"].index_information()
        ) == key_present

    def test_existing_index(self, orion_db):
        """Index should be added to database and reattempt should do nothing"""
        assert (
            "new_field_1" not in get_db(orion_db)["new_collection"].index_information()
        )

        orion_db.ensure_index("new_collection", "new_field", unique=True)
        assert "new_field_1" in get_db(orion_db)["new_collection"].index_information()

        # reattempt
        orion_db.ensure_index("new_collection", "new_field", unique=True)
        assert "new_field_1" in get_db(orion_db)["new_collection"].index_information()

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": ("end_time", "end_time_1"),
                "mongodb": ("end_time", "end_time_-1"),
            }
        ],
        indirect=True,
    )
    def test_ordered_index(self, orion_db, db_test_data):
        """Sort order should only be added to index when executed on a mongo
        database"""
        key, stored_key = db_test_data
        assert stored_key not in get_db(orion_db)["new_collection"].index_information()
        orion_db.ensure_index(
            "new_collection", [(key, Database.DESCENDING)], unique=True
        )
        assert stored_key in get_db(orion_db)["new_collection"].index_information()

    def test_compound_index(self, orion_db):
        """Tuple of Index should be added as a compound index."""
        assert (
            "name_1_metadata.user_1"
            not in get_db(orion_db)["experiments"].index_information()
        )
        orion_db.ensure_index(
            "experiments",
            [("name", Database.ASCENDING), ("metadata.user", Database.ASCENDING)],
            unique=True,
        )
        assert (
            "name_1_metadata.user_1"
            in get_db(orion_db)["experiments"].index_information()
        )


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestRead(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.read`."""

    def test_read_entries(self, orion_db, test_collection):
        """Fetch a whole entries."""
        loaded_config = orion_db.read(
            "test_collection", {"field1": "same1", "same_field": "same"}
        )
        assert loaded_config == [test_collection[0], test_collection[2]]

        loaded_config = orion_db.read(
            "test_collection", {"field1": "same1", "unique_field": "unique2"}
        )
        assert loaded_config == test_collection[2:]
        assert loaded_config[0]["_id"] == test_collection[2]["_id"]

    def test_read_with_id(self, orion_db, test_collection):
        """Query using ``_id`` key."""
        loaded_config = orion_db.read("test_collection", {"_id": 1})
        assert loaded_config == test_collection[1:2]

    def test_read_default(self, orion_db, test_collection):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            "test_collection",
            {"field1": "same1", "same_comp.ound": "same_compound"},
            selection={"unique_comp": 1, "_id": 0},
        )
        assert value == [
            {"unique_comp": test_collection[0]["unique_comp"]},
            {"unique_comp": test_collection[2]["unique_comp"]},
        ]

    def test_read_nothing(self, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            "experiments",
            {"name": "not_found", "metadata.user": "tsirif"},
            selection={"algorithms": 1},
        )
        assert value == []

    def test_read_trials(self, orion_db, test_collection):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            "test_collection",
            {
                "same_field": "same",
                "datetime": {"$gte": datetime(2017, 11, 23, 0, 0, 0)},
            },
        )
        assert value == test_collection[1:]

        value = orion_db.read(
            "test_collection",
            {
                "same_field": "same",
                "datetime": {"$gt": datetime(2017, 11, 23, 0, 0, 0)},
            },
        )
        assert value == test_collection[2:]

    @pytest.mark.db_types_only(["ephemeraldb", "pickleddb"])
    def test_null_comp(self, orion_db):
        """Fetch value(s) from an entry."""
        all_values = orion_db.read(
            "test_collection",
            {
                "same_field": "same",
                "datetime": {"$gte": datetime(2017, 11, 1, 0, 0, 0)},
            },
        )

        db = get_db(orion_db)
        db["test_collection"]._documents[0]._data["datetime"] = None
        dump_db(orion_db, db)

        values = orion_db.read(
            "test_collection",
            {
                "same_field": "same",
                "datetime": {"$gte": datetime(2017, 11, 1, 0, 0, 0)},
            },
        )
        assert len(values) == len(all_values) - 1


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestWrite(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.write`."""

    def test_insert_one(self, orion_db):
        """Should insert a single new entry in the collection."""
        item = {"exp_name": "supernaekei", "user": "tsirif"}
        count_before = orion_db.count("experiments")
        # call interface
        assert orion_db.write("experiments", item) == 1
        assert orion_db.count("experiments") == count_before + 1
        value = get_db(orion_db)["experiments"].find({"exp_name": "supernaekei"})[0]
        assert value == item

    def test_insert_many(self, orion_db):
        """Should insert two new entry (as a list) in the collection."""
        item = [
            {"exp_name": "supernaekei2", "user": "tsirif"},
            {"exp_name": "supernaekei3", "user": "tsirif"},
        ]
        count_before = get_db(orion_db)["experiments"].count()
        # call interface
        assert orion_db.write("experiments", item) == 2
        database = get_db(orion_db)
        assert database["experiments"].count() == count_before + 2
        value = database["experiments"].find({"exp_name": "supernaekei2"})[0]
        assert value == item[0]
        value = database["experiments"].find({"exp_name": "supernaekei3"})[0]
        assert value == item[1]

    def test_update_many_default(self, orion_db):
        """Should match existing entries, and update some of their keys."""
        filt = {"field1": "same1"}
        count_before = orion_db.count("test_collection")
        count_query = orion_db.count("test_collection", filt)
        # call interface
        assert (
            orion_db.write("test_collection", {"same_field": "diff"}, filt)
            == count_query
        )
        database = get_db(orion_db)
        assert database["test_collection"].count() == count_before
        value = list(database["test_collection"].find({}))
        assert value[0]["same_field"] == "diff"
        assert value[1]["same_field"] == "same"
        assert value[2]["same_field"] == "diff"

    def test_update_with_id(self, orion_db, test_collection):
        """Query using ``_id`` key."""
        filt = {"_id": test_collection[1]["_id"]}
        count_before = orion_db.count("test_collection")
        # call interface
        assert orion_db.write("test_collection", {"same_field": "diff"}, filt) == 1
        database = get_db(orion_db)
        assert database["test_collection"].count() == count_before
        value = list(database["test_collection"].find())
        assert value[0]["same_field"] == "same"
        assert value[1]["same_field"] == "diff"
        assert value[2]["same_field"] == "same"

    def test_no_upsert(self, orion_db):
        """Query with a non-existent ``_id`` should no upsert something."""
        assert (
            orion_db.write("experiments", {"pool_size": 66}, {"_id": "lalalathisisnew"})
            == 0
        )


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestReadAndWrite(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.read_and_write`."""

    def test_read_and_write_one(self, orion_db, test_collection):
        """Should read and update a single entry in the collection."""
        # Make sure there is only one match
        documents = orion_db.read("test_collection", {"unique_field": "unique1"})
        assert len(documents) == 1

        # Find and update atomically
        loaded_config = orion_db.read_and_write(
            "test_collection", {"unique_field": "unique1"}, {"field0": "lalala"}
        )
        test_collection[1]["field0"] = "lalala"
        assert loaded_config == test_collection[1]

    def test_read_and_write_many(self, orion_db, test_collection):
        """Should update only one entry."""
        documents = orion_db.read("test_collection", {"same_field": "same"})
        assert len(documents) > 1

        # Find many and update first one only
        loaded_config = orion_db.read_and_write(
            "test_collection", {"same_field": "same"}, {"unique_field": "lalala"}
        )

        test_collection[0]["unique_field"] = "lalala"
        assert loaded_config == test_collection[0]

        # Make sure it only changed the first document found
        documents = orion_db.read("test_collection", {"same_field": "same"})
        assert documents[0]["unique_field"] == "lalala"
        assert documents[1]["unique_field"] != "lalala"

    def test_read_and_write_no_match(self, orion_db):
        """Should return None when there is no match."""
        loaded_config = orion_db.read_and_write(
            "experiments", {"name": "lalala"}, {"pool_size": "lalala"}
        )

        assert loaded_config is None


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestRemove(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.remove`."""

    def test_remove_many_default(self, orion_db, test_collection):
        """Should match existing entries, and delete them all."""
        filt = {"field1": "same1"}
        database = get_db(orion_db)
        count_before = database["test_collection"].count()
        count_filt = database["test_collection"].count(filt)
        # call interface
        assert orion_db.remove("test_collection", filt) == count_filt
        database = get_db(orion_db)
        assert database["test_collection"].count() == count_before - count_filt
        assert database["test_collection"].count() == 1
        loaded_config = list(database["test_collection"].find())
        assert loaded_config == test_collection[1:2]

    def test_remove_with_id(self, orion_db, test_collection):
        """Query using ``_id`` key."""
        filt = {"_id": test_collection[0]["_id"]}

        database = get_db(orion_db)
        count_before = database["test_collection"].count()
        # call interface
        assert orion_db.remove("test_collection", filt) == 1
        database = get_db(orion_db)
        assert database["test_collection"].count() == count_before - 1
        loaded_configs = list(database["test_collection"].find())
        assert loaded_configs == test_collection[1:]

    def test_remove_update_indexes(self, orion_db, test_collection):
        """Verify that indexes are properly update after deletion."""
        with pytest.raises(DuplicateKeyError):
            orion_db.write("test_collection", {"_id": test_collection[0]["_id"]})
        with pytest.raises(DuplicateKeyError):
            orion_db.write("test_collection", {"_id": test_collection[1]["_id"]})

        filt = {"_id": test_collection[0]["_id"]}

        database = get_db(orion_db)
        count_before = database["test_collection"].count()
        # call interface
        assert orion_db.remove("test_collection", filt) == 1
        database = get_db(orion_db)
        assert database["test_collection"].count() == count_before - 1
        # Should not fail now, otherwise it means the indexes were not updated properly during
        # remove()
        orion_db.write("test_collection", filt)
        # And this should still fail
        with pytest.raises(DuplicateKeyError):
            orion_db.write("test_collection", {"_id": test_collection[1]["_id"]})


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestCount(object):
    """Calls :meth:`orion.core.io.database.AbstractDB.count`."""

    def test_count_default(self, orion_db, test_collection):
        """Call just with collection name."""
        found = orion_db.count("test_collection")
        assert found == len(test_collection)

    def test_count_query(self, orion_db):
        """Call with a query."""
        found = orion_db.count("test_collection", {"field1": "same1"})
        assert found == 2

    def test_count_query_with_id(self, orion_db, test_collection):
        """Call querying with unique _id."""
        found = orion_db.count("test_collection", {"_id": test_collection[2]["_id"]})
        assert found == 1

    def test_count_nothing(self, orion_db, test_collection):
        """Call with argument that will not find anything."""
        found = orion_db.count("test_collection", {"name": "lalalanotfound"})
        assert found == 0


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestIndexInformation(object):
    """Calls :meth:`orion.core.io.database.AbstractDB.index_information`."""

    def test_no_index(self, orion_db):
        """Test that no index is returned when there is none."""
        assert orion_db.index_information("test_collection") == {"_id_": True}

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": (False, "name", {"_id_": True}),
                "mongodb": (False, "name", {"_id_": True, "name_1": False}),
            },
            {
                "ephemeraldb_pickleddb": (True, "name", {"_id_": True, "name_1": True}),
                "mongodb": (True, "name", {"_id_": True, "name_1": True}),
            },
        ],
        indirect=True,
    )
    def test_single_index(self, orion_db, db_test_data):
        """Test that single indexes are ignored if not unique."""
        unique, key, index_information = db_test_data
        orion_db.ensure_index("experiments", [(key, Database.ASCENDING)], unique=unique)

        assert orion_db.index_information("experiments") == index_information

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": (False, "name", {"_id_": True}),
                "mongodb": (False, "name", {"_id_": True, "name_-1": False}),
            },
            {
                "ephemeraldb_pickleddb": (True, "name", {"_id_": True, "name_1": True}),
                "mongodb": (True, "name", {"_id_": True, "name_-1": True}),
            },
        ],
        indirect=True,
    )
    def test_ordered_index(self, orion_db, db_test_data):
        """Test that ordered indexes are not taken into account."""
        unique, key, index_information = db_test_data
        orion_db.ensure_index(
            "experiments", [(key, Database.DESCENDING)], unique=unique
        )

        assert orion_db.index_information("experiments") == index_information

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": (
                    False,
                    [("name", Database.DESCENDING), ("version", Database.ASCENDING)],
                    {"_id_": True},
                ),
                "mongodb": (
                    False,
                    [("name", Database.DESCENDING), ("version", Database.ASCENDING)],
                    {"_id_": True, "name_-1_version_1": False},
                ),
            },
            {
                "ephemeraldb_pickleddb": (
                    True,
                    [("name", Database.DESCENDING), ("version", Database.ASCENDING)],
                    {"_id_": True, "name_1_version_1": True},
                ),
                "mongodb": (
                    True,
                    [("name", Database.DESCENDING), ("version", Database.ASCENDING)],
                    {"_id_": True, "name_-1_version_1": True},
                ),
            },
        ],
        indirect=True,
    )
    def test_compound_index(self, orion_db, db_test_data):
        """Test representation of compound indexes."""
        unique, keys_pair, index_information = db_test_data
        orion_db.ensure_index(
            "experiments",
            keys_pair,
            unique=unique,
        )

        assert orion_db.index_information("experiments") == index_information


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestDropIndex(object):
    """Calls :meth:`orion.core.io.database.AbstractDB.drop_index`."""

    def test_no_index(self, orion_db):
        """Test that no index is returned when there is none."""
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index("test_collection", "i_dont_exist")
        assert "index not found with name" in str(exc.value)

    def test_drop_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index(
            "experiments", [("name", Database.ASCENDING)], unique=True
        )
        assert orion_db.index_information("experiments") == {
            "_id_": True,
            "name_1": True,
        }
        orion_db.drop_index("experiments", "name_1")
        assert orion_db.index_information("experiments") == {"_id_": True}

    @pytest.mark.db_types_only(["ephemeraldb", "pickleddb"])
    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "ephemeraldb_pickleddb": (
                    [("name", Database.DESCENDING)],
                    "name_1",
                    {"_id_": True, "name_1": True},
                ),
            },
            {
                "ephemeraldb_pickleddb": (
                    [("name", Database.ASCENDING), ("version", Database.DESCENDING)],
                    "name_1_version_1",
                    {"_id_": True, "name_1_version_1": True},
                ),
            },
        ],
        indirect=True,
    )
    def test_drop_ordered_index(self, orion_db, db_test_data):
        """Test with single and compound indexes."""
        keys, stored_keys, index_information = db_test_data
        orion_db.ensure_index("experiments", keys, unique=True)
        assert orion_db.index_information("experiments") == index_information
        orion_db.drop_index("experiments", stored_keys)
        assert orion_db.index_information("experiments") == {"_id_": True}


@insert_test_collection
def test_serializable(orion_db):
    serialized = pickle.dumps(orion_db)
    deserialized = pickle.loads(serialized)
    assert len(orion_db.read("test_collection", {})) > 0
    assert orion_db.read("test_collection", {}) == deserialized.read(
        "test_collection", {}
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for client helper functions."""
import os

import pytest

import orion.core.cli
from orion.core.utils.exceptions import InvalidResult
from orion.core.worker.consumer import Consumer


def test_interrupt(storage, monkeypatch, capsys):
    """Test interruption from within user script."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    error_code = orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "1",
            "--worker-trials",
            "1",
            "python",
            "black_box.py",
            "interrupt_trial",
        ]
        + user_args
    )

    assert error_code == 130

    captured = capsys.readouterr()
    assert captured.out == "Orion is interrupted.\n"
    assert captured.err == ""

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 1
    assert trials[0].status == "interrupted"


def test_interrupt_diff_code(storage, monkeypatch, capsys):
    """Test interruption from within user script with custom int code"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    # Set local to 200
    orion.core.config.worker.interrupt_signal_code = 200

    # But child won't be passed ORION_INTERRUPT_CODE and therefore will send default code 130
    def empty_env(self, trial, results_file=None):
        return os.environ

    with monkeypatch.context() as m:
        m.setattr(Consumer, "get_execution_environment", empty_env)

        # Interrupt won't be interpreted properly and trials will be marked as broken
        error_code = orion.core.cli.main(
            [
                "hunt",
                "--config",
                "./orion_config.yaml",
                "--worker-max-broken",
                "2",
                "--exp-max-trials",
                "2",
                "--worker-trials",
                "2",
                "python",
                "black_box.py",
                "interrupt_trial",
            ]
            + user_args
        )

        assert error_code == 0

        exp = list(storage.fetch_experiments({"name": "voila_voici"}))
        exp = exp[0]
        exp_id = exp["_id"]
        trials = list(storage.fetch_trials(uid=exp_id))
        assert len(trials) == 2
        assert trials[0].status == "broken"

    # This time we use true `get_execution_environment which pass properly int code to child.
    error_code = orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trial",
            "1",
            "--worker-trials",
            "1",
            "python",
            "black_box.py",
            "interrupt_trial",
        ]
        + user_args
    )

    assert error_code == 130

    captured = capsys.readouterr()
    assert "Orion is interrupted.\n" in captured.out
    assert captured.err == ""

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 3
    assert trials[-1].status == "interrupted"


# TODO:

# test no call to any report

# Add all this in DOC.


@pytest.mark.parametrize("fct", ["report_bad_trial", "report_objective"])
def test_report_no_name(storage, monkeypatch, fct):
    """Test report helper functions with default names"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "2",
            "--worker-trials",
            "2",
            "python",
            "black_box.py",
            fct,
            "--objective",
            "1.0",
        ]
        + user_args
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 2
    assert trials[0].status == "completed"
    assert trials[0].results[0].name == "objective"
    assert trials[0].results[0].type == "objective"
    assert trials[0].results[0].value == 1.0


@pytest.mark.parametrize("fct", ["report_bad_trial", "report_objective"])
def test_report_with_name(storage, monkeypatch, fct):
    """Test report helper functions with custom names"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "2",
            "--worker-trials",
            "2",
            "python",
            "black_box.py",
            fct,
            "--objective",
            "1.0",
            "--name",
            "metric",
        ]
        + user_args
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 2
    assert trials[0].status == "completed"
    assert trials[0].results[0].name == "metric"
    assert trials[0].results[0].type == "objective"
    assert trials[0].results[0].value == 1.0


@pytest.mark.parametrize("fct", ["report_bad_trial", "report_objective"])
def test_report_with_bad_objective(storage, monkeypatch, fct):
    """Test report helper functions with bad objective types"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    with pytest.raises(InvalidResult) as exc:
        orion.core.cli.main(
            [
                "hunt",
                "--config",
                "./orion_config.yaml",
                "--exp-max-trials",
                "2",
                "--worker-trials",
                "2",
                "python",
                "black_box.py",
                fct,
                "--objective",
                "oh oh",
            ]
            + user_args
        )

    assert "must contain a type `objective` with type float/int" in str(exc.value)


def test_report_with_bad_trial_no_objective(storage, monkeypatch):
    """Test bad trial report helper function with default objective."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "2",
            "--worker-trials",
            "2",
            "python",
            "black_box.py",
            "report_bad_trial",
        ]
        + user_args
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 2
    assert trials[0].status == "completed"
    assert trials[0].results[0].name == "objective"
    assert trials[0].results[0].type == "objective"
    assert trials[0].results[0].value == 1e10


def test_report_with_bad_trial_with_data(storage, monkeypatch):
    """Test bad trial report helper function with additional data."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "2",
            "--worker-trials",
            "2",
            "python",
            "black_box.py",
            "report_bad_trial",
            "--data",
            "another",
        ]
        + user_args
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 2
    assert trials[0].status == "completed"
    assert trials[0].results[0].name == "objective"
    assert trials[0].results[0].type == "objective"
    assert trials[0].results[0].value == 1e10

    assert trials[0].results[1].name == "another"
    assert trials[0].results[1].type == "constraint"
    assert trials[0].results[1].value == 1.0


def test_no_report(storage, monkeypatch, capsys):
    """Test script call without any results reported."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    errorcode = orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--exp-max-trials",
            "2",
            "--worker-trials",
            "2",
            "python",
            "black_box.py",
            "no_report",
        ]
        + user_args
    )

    assert errorcode == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Cannot parse result file" in captured.err



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures and utils for benchmark unit tests."""

import pytest

from orion.benchmark.assessment import AverageRank, AverageResult
from orion.benchmark.task import CarromTable, RosenBrock
from orion.testing import generate_trials


@pytest.fixture()
def task_number():
    """Return task number for assessment"""
    return 2


@pytest.fixture()
def max_trial():
    """Return max trials for task"""
    return 3


@pytest.fixture
def benchmark_algorithms():
    """Return a list of algorithms suitable for Orion experiment"""
    return [{"algorithm": {"random": {"seed": 1}}}, {"algorithm": {"tpe": {"seed": 1}}}]


@pytest.fixture
def benchmark_config(benchmark_algorithms):
    config = {
        "name": "bm00001",
        "algorithms": benchmark_algorithms,
        "targets": [
            {
                "assess": {
                    "AverageResult": {"task_num": 2},
                    "AverageRank": {"task_num": 2},
                },
                "task": {
                    "RosenBrock": {"dim": 3, "max_trials": 25},
                    "CarromTable": {"max_trials": 20},
                },
            }
        ],
    }

    return config


@pytest.fixture
def benchmark_config_py(benchmark_algorithms):
    config = dict(
        name="bm00001",
        algorithms=benchmark_algorithms,
        targets=[
            {
                "assess": [AverageResult(2), AverageRank(2)],
                "task": [RosenBrock(25, dim=3), CarromTable(20)],
            }
        ],
    )
    return config


@pytest.fixture()
def experiment_config(max_trial, benchmark_algorithms):
    """Return a experiment template configure"""
    config = dict(
        name="experiment-name",
        space={"x": "uniform(0, 200)"},
        metadata={
            "user": "test-user",
            "orion_version": "XYZ",
            "VCS": {
                "type": "git",
                "is_dirty": False,
                "HEAD_sha": "test",
                "active_branch": None,
                "diff_sha": "diff",
            },
        },
        version=1,
        pool_size=1,
        max_trials=max_trial,
        working_dir="",
        algorithms=benchmark_algorithms[0]["algorithm"],
        producer={"strategy": "NoParallelStrategy"},
    )
    return config


@pytest.fixture()
def trial_config():
    """Return a trial template configure"""
    trial_config = {
        "experiment": 0,
        "status": "completed",
        "worker": None,
        "start_time": None,
        "end_time": None,
        "heartbeat": None,
        "results": [],
        "params": [],
    }
    return trial_config


@pytest.fixture
def study_experiments_config(
    experiment_config, trial_config, benchmark_algorithms, task_number, max_trial
):
    config = dict(
        exp_config=experiment_config,
        trial_config=trial_config,
        algorithms=benchmark_algorithms,
        task_number=task_number,
        max_trial=max_trial,
    )

    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.transformer`."""
import copy
import itertools
from collections import OrderedDict

import numpy
import pytest

from orion.algo.space import Categorical, Dimension, Integer, Real, Space
from orion.core.worker.transformer import (
    Compose,
    Enumerate,
    Identity,
    Linearize,
    OneHotEncode,
    Precision,
    Quantize,
    ReshapedDimension,
    ReshapedSpace,
    Reverse,
    TransformedDimension,
    TransformedSpace,
    View,
    build_required_space,
)


class TestIdentity(object):
    """Test subclasses of `Identity` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Identity()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Identity()
        assert t.domain_type is None
        assert t.target_type is None

        t = Identity("mpogias")
        assert t.domain_type == "mpogias"
        assert t.target_type == "mpogias"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Identity()
        assert t.transform("yo") == "yo"

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Identity()
        assert t.reverse("yo") == "yo"

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Identity()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Identity()
        assert t.repr_format("asfa") == "asfa"


class TestReverse(object):
    """Test subclasses of `Reverse` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Reverse(Quantize())
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Reverse(Quantize())
        assert t.domain_type == "integer"
        assert t.target_type == "real"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Reverse(Quantize())
        assert t.transform(9) == 9.0
        assert t.transform(5) == 5.0
        assert numpy.all(t.transform([9, 5]) == numpy.array([9.0, 5.0], dtype=float))

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Reverse(Quantize())
        assert t.reverse(8.6) == 8
        assert t.reverse(5.3) == 5
        assert numpy.all(t.reverse([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Reverse(Quantize())
        assert t.infer_target_shape((5,)) == (5,)

    def test_no_reverse_one_hot_encode(self):
        """Do NOT support real to categorical."""
        with pytest.raises(AssertionError):
            Reverse(OneHotEncode([1, 2, 3]))

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Reverse(Quantize())
        assert t.repr_format("asfa") == "ReverseQuantize(asfa)"


class TestCompose(object):
    """Test subclasses of `Compose` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Compose([])
        assert t.domain_type is None
        assert t.target_type is None

        t = Compose([], "real")
        assert t.domain_type == "real"
        assert t.target_type == "real"

        t = Compose([Quantize()], "real")
        assert t.domain_type == "real"
        assert t.target_type == "integer"

        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        assert t.domain_type == "categorical"
        assert t.target_type == "real"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        assert numpy.all(t.transform(2) == numpy.array((1.0, 0.0, 0.0)))
        assert numpy.all(t.transform("asfa") == numpy.array((0.0, 1.0, 0.0)))
        assert numpy.all(t.transform("ipsi") == numpy.array((0.0, 0.0, 1.0)))
        with pytest.raises(KeyError):
            t.transform("aafdasfa")
        assert numpy.all(
            t.transform([["ipsi", "asfa"], [2, "ipsi"]])
            == numpy.array(
                [[(0.0, 0.0, 1.0), (0.0, 1.0, 0.0)], [(1.0, 0.0, 0.0), (0.0, 0.0, 1.0)]]
            )
        )

        t = Compose([Enumerate([2, "asfa"]), OneHotEncode(2)], "categorical")
        assert t.transform(2) == 0.0
        assert t.transform("asfa") == 1.0
        with pytest.raises(KeyError):
            t.transform("ipsi")
        assert numpy.all(
            t.transform([["asfa", "asfa"], [2, "asfa"]])
            == numpy.array([[1.0, 1.0], [0.0, 1.0]])
        )

        # for the crazy enough
        t = Compose([Enumerate([2]), OneHotEncode(1)], "categorical")
        assert t.transform(2) == 0.0
        with pytest.raises(KeyError):
            t.transform("ipsi")
        assert numpy.all(t.transform([[2, 2], [2, 2]]) == [[0, 0], [0, 0]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        assert t.reverse((0.9, 0.8, 0.3)) == 2
        assert t.reverse((-0.3, 2.0, 0.0)) == "asfa"
        assert t.reverse((0.0, 0.0, 1.0)) == "ipsi"
        with pytest.raises(AssertionError):
            t.reverse((0.0, 0.0, 0.0, 1.0))
        assert numpy.all(
            t.reverse(
                numpy.array(
                    [
                        [(0.0, 0.0, 1.0), (0.0, 1.0, 0.0)],
                        [(1.0, 0.0, 0.0), (0.0, 0.0, 1.0)],
                    ]
                )
            )
            == numpy.array([["ipsi", "asfa"], [2, "ipsi"]], dtype=numpy.object)
        )

        t = Compose([Enumerate([2, "asfa"]), OneHotEncode(2)], "categorical")
        assert t.reverse(0.3) == 2
        assert t.reverse(2.0) == "asfa"
        assert numpy.all(
            t.reverse((0.0, 0.0, 0.0, 1.0))
            == numpy.array([2, 2, 2, "asfa"], dtype=numpy.object)
        )
        assert numpy.all(
            t.reverse(numpy.array([[0.55, 3.0], [-0.6, 1.0]]))
            == numpy.array([["asfa", "asfa"], [2, "asfa"]], dtype=numpy.object)
        )

        # for the crazy enough
        t = Compose([Enumerate([2]), OneHotEncode(1)], "categorical")
        assert t.reverse(0) == 2
        assert t.reverse(5.0) == 2
        assert t.reverse(0.2) == 2
        assert t.reverse(-0.2) == 2
        assert numpy.all(
            t.reverse([[0.5, 0], [1.0, 55]])
            == numpy.array([[2, 2], [2, 2]], dtype=numpy.object)
        )

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        assert t.infer_target_shape((2, 5)) == (2, 5, 3)

        t = Compose([Enumerate([2, "asfa"]), OneHotEncode(2)], "categorical")
        assert t.infer_target_shape((2, 5)) == (2, 5)

        t = Compose([Enumerate([2]), OneHotEncode(1)], "categorical")
        assert t.infer_target_shape((2, 5)) == (2, 5)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Compose([Enumerate([2, "asfa", "ipsi"]), OneHotEncode(3)], "categorical")
        assert t.repr_format("asfa") == "OneHotEncode(Enumerate(asfa))"


class TestPrecision(object):
    """Test subclasses of `Precision` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Precision()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Precision()
        assert t.domain_type == "real"
        assert t.target_type == "real"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Precision(precision=4)
        assert t.transform(8.654321098) == 8.654
        assert t.transform(0.000123456789) == 0.0001235
        assert numpy.all(
            t.transform([8.654321098, 0.000123456789])
            == numpy.array([8.654, 0.0001235], dtype=float)
        )

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Precision()
        assert t.reverse(9.0) == 9.0
        assert t.reverse(5.0) == 5.0
        assert numpy.all(t.reverse([9.0, 5.0]) == numpy.array([9.0, 5.0], dtype=float))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Precision()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Precision()
        assert t.repr_format("asfa") == "Precision(4, asfa)"


class TestQuantize(object):
    """Test subclasses of `Quantize` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Quantize()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Quantize()
        assert t.domain_type == "real"
        assert t.target_type == "integer"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Quantize()
        assert t.transform(8.6) == 8
        assert t.transform(5.3) == 5
        assert numpy.all(t.transform([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Quantize()
        assert t.reverse(9) == 9.0
        assert t.reverse(5) == 5.0
        assert numpy.all(t.reverse([9, 5]) == numpy.array([9.0, 5.0], dtype=float))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Quantize()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Quantize()
        assert t.repr_format("asfa") == "Quantize(asfa)"


class TestEnumerate(object):
    """Test subclasses of `Enumerate` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Enumerate([2, "asfa", "ipsi"])
        # Copy won't fail if vectorized function is not called at least once.
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Enumerate([2, "asfa", "ipsi"])
        assert t.domain_type == "categorical"
        assert t.target_type == "integer"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Enumerate([2, "asfa", "ipsi"])
        assert t.transform(2) == 0
        assert t.transform("asfa") == 1
        assert t.transform("ipsi") == 2
        with pytest.raises(KeyError):
            t.transform("aafdasfa")
        assert numpy.all(
            t.transform([["ipsi", "asfa"], [2, "ipsi"]]) == [[2, 1], [0, 2]]
        )

        # for the crazy enough
        t = Enumerate([2])
        assert t.transform(2) == 0
        with pytest.raises(KeyError):
            t.transform("aafdasfa")
        assert numpy.all(t.transform([[2, 2], [2, 2]]) == [[0, 0], [0, 0]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Enumerate([2, "asfa", "ipsi"])
        assert t.reverse(0) == 2
        assert t.reverse(1) == "asfa"
        assert t.reverse(2) == "ipsi"
        with pytest.raises(IndexError):
            t.reverse(3)
        assert numpy.all(
            t.reverse([[2, 1], [0, 2]])
            == numpy.array([["ipsi", "asfa"], [2, "ipsi"]], dtype=numpy.object)
        )

        # for the crazy enough
        t = Enumerate([2])
        assert t.reverse(0) == 2
        with pytest.raises(IndexError):
            t.reverse(1)
        assert numpy.all(
            t.reverse([[0, 0], [0, 0]])
            == numpy.array([[2, 2], [2, 2]], dtype=numpy.object)
        )

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Enumerate([2, "asfa", "ipsi"])
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Enumerate([2, "asfa", "ipsi"])
        assert t.repr_format("asfa") == "Enumerate(asfa)"


class TestOneHotEncode(object):
    """Test subclasses of `OneHotEncode` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = OneHotEncode(3)
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = OneHotEncode(3)
        assert t.domain_type == "integer"
        assert t.target_type == "real"

    def test_transform(self):
        """Check if it transforms properly."""
        t = OneHotEncode(3)
        assert numpy.all(t.transform(0) == numpy.array((1.0, 0.0, 0.0)))
        assert numpy.all(t.transform(1) == numpy.array((0.0, 1.0, 0.0)))
        assert numpy.all(t.transform(2) == numpy.array((0.0, 0.0, 1.0)))
        with pytest.raises(AssertionError):
            t.transform(4)
        with pytest.raises(AssertionError):
            t.transform(-1)
        with pytest.raises(AssertionError):
            t.transform(2.2)
        assert numpy.all(
            t.transform([[2, 1], [0, 2]])
            == numpy.array(
                [[(0.0, 0.0, 1.0), (0.0, 1.0, 0.0)], [(1.0, 0.0, 0.0), (0.0, 0.0, 1.0)]]
            )
        )

        t = OneHotEncode(2)
        assert t.transform(0) == 0.0
        assert t.transform(1) == 1.0
        with pytest.raises(TypeError):
            t.transform("ipsi")
        assert numpy.all(
            t.transform([[1, 1], [0, 1]]) == numpy.array([[1.0, 1.0], [0.0, 1.0]])
        )

        # for the crazy enough
        t = OneHotEncode(1)
        assert t.transform(0) == 0.0
        with pytest.raises(TypeError):
            t.transform("ipsi")
        assert numpy.all(t.transform([[0, 0], [0, 0]]) == [[0.0, 0.0], [0.0, 0.0]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = OneHotEncode(3)
        assert t.reverse((0.9, 0.8, 0.3)) == 0
        assert t.reverse((-0.3, 2.0, 0.0)) == 1
        assert t.reverse((0.0, 0.0, 1.0)) == 2
        with pytest.raises(AssertionError):
            t.reverse((0.0, 0.0, 0.0, 1.0))
        assert numpy.all(
            t.reverse(
                numpy.array(
                    [
                        [[0.0, 0.0, 1.0], [0.0, 1.0, 0.0]],
                        [[1.0, 0.0, 0.0], [0.0, 0.0, 1.0]],
                    ]
                )
            )
            == numpy.array([[2, 1], [0, 2]], dtype=int)
        )

        t = OneHotEncode(2)
        assert t.reverse(0.3) == 0
        assert t.reverse(2.0) == 1
        assert numpy.all(
            t.reverse((0.0, 0.0, 0.0, 1.0)) == numpy.array([0, 0, 0, 1], dtype=int)
        )
        assert numpy.all(
            t.reverse(numpy.array([[0.55, 3.0], [-0.6, 1.0]]))
            == numpy.array([[1, 1], [0, 1]], dtype=int)
        )

        # for the crazy enough
        t = OneHotEncode(1)
        assert t.reverse(0) == 0
        assert t.reverse(5.0) == 0
        assert t.reverse(0.2) == 0
        assert t.reverse(-0.2) == 0
        assert numpy.all(
            t.reverse([[0.5, 0], [1.0, 55]]) == numpy.array([[0, 0], [0, 0]], dtype=int)
        )

    def test_interval(self):
        """Test that the onehot interval has the proper dimensions"""
        t = OneHotEncode(3)
        low, high = t.interval()
        assert (low == numpy.zeros(3)).all()
        assert (high == numpy.ones(3)).all()

        t = OneHotEncode(2)
        low, high = t.interval()
        assert (low == numpy.zeros(1)).all()
        assert (high == numpy.ones(1)).all()

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = OneHotEncode(3)
        assert t.infer_target_shape((2, 5)) == (2, 5, 3)

        t = OneHotEncode(2)
        assert t.infer_target_shape((2, 5)) == (2, 5)

        t = OneHotEncode(1)
        assert t.infer_target_shape((2, 5)) == (2, 5)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = OneHotEncode(3)
        assert t.repr_format("asfa") == "OneHotEncode(asfa)"


class TestLinearize(object):
    """Test subclasses of `Linearize` transformation."""

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Linearize()
        assert t.domain_type == "real"
        assert t.target_type == "real"

    def test_transform(self):
        """Check if it transforms properly."""
        t = Linearize()
        assert t.transform(numpy.e) == 1
        t.transform(0)

    def test_reverse(self):
        """Check if it reverses `transform` properly."""
        t = Linearize()
        assert t.reverse(1) == numpy.e

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Linearize()
        assert t.repr_format(1.0) == "Linearize(1.0)"


class TestView(object):
    """Test subclasses of `View` transformation."""

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do what's expected."""
        t = View(shape=None, index=None, domain_type="some fancy type")
        assert t.domain_type == "some fancy type"
        assert t.target_type == "some fancy type"

    def test_transform(self):
        """Check if it transforms properly."""
        shape = (3, 4, 5)
        index = (0, 2, 1)
        t = View(shape=shape, index=index)
        a = numpy.zeros(shape)
        a[index] = 2
        assert t.transform(a) == 2

    def test_reverse(self):
        """Check if it reverses `transform` properly."""
        shape = (3, 4, 5)
        index = (0, 2, 1)
        a = numpy.zeros(shape)
        a[index] = 2
        flattened = a.reshape(-1).tolist()
        point = [None] + flattened + [None]
        t = View(shape=shape, index=(0, 0, 0))
        numpy.testing.assert_equal(t.reverse(point, 1), a)

    def test_first(self):
        """Test that views are correctly identified as first"""
        shape = (3, 4, 5)
        assert View(shape=shape, index=(0, 0, 0)).first
        assert not View(shape=shape, index=(0, 1, 0)).first

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        shape = (3, 4, 5)
        index = (0, 2, 1)
        t = View(shape=shape, index=index)
        assert t.repr_format(1.0) == "View(shape=(3, 4, 5), index=(0, 2, 1), 1.0)"


@pytest.fixture(scope="module")
def dim():
    """Create an example of `Dimension`."""
    dim = Real("yolo0", "norm", 0.9, shape=(3, 2))
    return dim


@pytest.fixture(scope="module")
def logdim():
    """Create an log example of `Dimension`."""
    dim = Real("yolo4", "reciprocal", 1.0, 10.0, shape=(3, 2))
    return dim


@pytest.fixture(scope="module")
def logintdim():
    """Create an log integer example of `Dimension`."""
    dim = Integer("yolo5", "reciprocal", 1, 10, shape=(3, 2))
    return dim


@pytest.fixture(scope="module")
def tdim(dim):
    """Create an example of `TransformedDimension`."""
    transformers = [Quantize()]
    tdim = TransformedDimension(Compose(transformers, dim.type), dim)
    return tdim


@pytest.fixture(scope="module")
def rdims(tdim):
    """Create an example of `ReshapedDimension`."""
    transformations = {}
    for index in itertools.product(*map(range, tdim.shape)):
        key = f'{tdim.name}[{",".join(map(str, index))}]'
        transformations[key] = ReshapedDimension(
            transformer=View(tdim.shape, index, tdim.type),
            original_dimension=tdim,
            name=key,
            index=0,
        )

    return transformations


@pytest.fixture(scope="module")
def rdim(dim, rdims):
    """Single ReshapedDimension"""
    return rdims[f"{dim.name}[0,1]"]


@pytest.fixture(scope="module")
def dim2():
    """Create a second example of `Dimension`."""
    probs = (0.1, 0.2, 0.3, 0.4)
    categories = ("asdfa", "2", "3", "4")
    categories = OrderedDict(zip(categories, probs))
    dim2 = Categorical("yolo2", categories)
    return dim2


@pytest.fixture(scope="module")
def tdim2(dim2):
    """Create a second example of `TransformedDimension`."""
    transformers = [Enumerate(dim2.categories), OneHotEncode(len(dim2.categories))]
    tdim2 = TransformedDimension(Compose(transformers, dim2.type), dim2)
    return tdim2


@pytest.fixture(scope="module")
def rdims2(tdim2):
    """Create a categorical example of `ReshapedDimension`."""
    transformations = {}
    for index in itertools.product(*map(range, tdim2.shape)):
        key = f'{tdim2.name}[{",".join(map(str, index))}]'
        transformations[key] = ReshapedDimension(
            transformer=View(tdim2.shape, index, tdim2.type),
            original_dimension=tdim2,
            name=key,
            index=1,
        )

    return transformations


@pytest.fixture(scope="module")
def rdim2(dim2, rdims2):
    """Single ReshapedDimension"""
    return rdims2[f"{dim2.name}[1]"]


@pytest.fixture(scope="module")
def dim3():
    """Create an example of integer `Dimension`."""
    return Integer("yolo3", "uniform", 3, 7)


@pytest.fixture(scope="module")
def tdim3(dim3):
    """Create an example of integer `Dimension`."""
    return TransformedDimension(Compose([], dim3.type), dim3)


@pytest.fixture(scope="module")
def rdims3(tdim3):
    """Create an example of integer `Dimension`."""
    rdim3 = ReshapedDimension(
        transformer=Identity(tdim3.type), original_dimension=tdim3, index=2
    )

    return {tdim3.name: rdim3}


class TestTransformedDimension(object):
    """Check functionality of class `TransformedDimension`."""

    def test_transform(self, tdim):
        """Check method `transform`."""
        assert tdim.transform(8.6) == 8
        assert tdim.transform(5.3) == 5
        assert numpy.all(tdim.transform([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_reverse(self, tdim):
        """Check method `reverse`."""
        assert tdim.reverse(9) == 9.0
        assert tdim.reverse(5) == 5.0
        assert numpy.all(tdim.reverse([9, 5]) == numpy.array([9.0, 5.0], dtype=float))

    def test_interval(self, tdim):
        """Check method `interval`."""
        tmp1 = tdim.original_dimension._low
        tmp2 = tdim.original_dimension._high
        tdim.original_dimension._low = -0.6
        tdim.original_dimension._high = 1.2

        assert tdim.interval() == (-1, 1)

        tdim.original_dimension._low = tmp1
        tdim.original_dimension._high = tmp2

    def test_interval_from_categorical(self, tdim2):
        """Check how we should treat interval when original dimension is categorical."""
        low, high = tdim2.interval()
        assert (low == numpy.zeros(4)).all()
        assert (high == numpy.ones(4)).all()

    def test_contains(self, tdim):
        """Check method `__contains__`."""
        assert [[1, 1], [3, 1], [1, 2]] in tdim

        tmp1 = tdim.original_dimension._low
        tmp2 = tdim.original_dimension._high
        tdim.original_dimension._low = -0.6
        tdim.original_dimension._high = 1.2

        assert [[1, 1], [3, 1], [1, 2]] not in tdim

        tdim.original_dimension._low = tmp1
        tdim.original_dimension._high = tmp2

    def test_contains_from_categorical(self, tdim2):
        """Check method `__contains__` when original is categorical."""
        assert (0, 0, 0, 1) in tdim2
        assert (0, 2, 0, 1) in tdim2
        assert (0, 2, 0) not in tdim2

    def test_eq(self, tdim, tdim2):
        """Return True if other is the same transformed dimension as self"""
        assert tdim != tdim2
        assert tdim == copy.deepcopy(tdim)

    def test_hash(self, tdim, tdim2):
        """Test that hash is consistent for identical and different transformed dimensions"""
        assert hash(tdim) != hash(tdim2)
        assert hash(tdim) == hash(copy.deepcopy(tdim))

    def test_get_hashable_members(self, tdim, tdim2):
        """Test that hashable members of the transformed dimensions are the aggregation of
        transformer's and original dimension's hashable members.
        """
        assert tdim._get_hashable_members() == (
            "Compose",
            "Quantize",
            "real",
            "integer",
            "Identity",
            "real",
            "real",
            "yolo0",
            (3, 2),
            "real",
            (0.9,),
            (),
            None,
            "norm",
        )
        assert tdim2._get_hashable_members() == (
            "Compose",
            "OneHotEncode",
            "integer",
            "real",
            4,
            "Compose",
            "Enumerate",
            "categorical",
            "integer",
            "Identity",
            "categorical",
            "categorical",
            "yolo2",
            (),
            "categorical",
            (),
            (),
            None,
            "Distribution",
        )

    def test_validate(self, tdim, tdim2):
        """Validate original_dimension"""
        # It pass
        tdim.validate()
        tdim2.validate()

        # We break it
        tdim.original_dimension._kwargs["size"] = (2,)
        tdim2.original_dimension._default_value = "bad-default"

        # It does not pass
        with pytest.raises(ValueError) as exc:
            tdim.validate()
        assert "Use 'shape' keyword only instead of 'size'." in str(exc.value)

        with pytest.raises(ValueError) as exc:
            tdim2.validate()
        assert "bad-default is not a valid value for this Dimension." in str(exc.value)

        tdim.original_dimension._kwargs.pop("size")
        tdim2.original_dimension._default_value = Dimension.NO_DEFAULT_VALUE

    def test_repr(self, tdim):
        """Check method `__repr__`."""
        assert (
            str(tdim)
            == "Quantize(Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None))"
        )  # noqa

    def test_name_property(self, tdim):
        """Check property `name`."""
        assert tdim.name == "yolo0"

    def test_type_property(self, tdim, tdim2):
        """Check property `type`."""
        assert tdim.type == "integer"
        assert tdim2.type == "real"

    def test_prior_name_property(self, tdim, tdim2):
        """Check property `prior_name`."""
        assert tdim.prior_name == "norm"
        assert tdim2.prior_name == "choices"

    def test_shape_property(self, tdim, tdim2):
        """Check property `shape`."""
        assert tdim.original_dimension.shape == (3, 2)
        assert tdim.shape == (3, 2)
        assert tdim2.original_dimension.shape == ()
        assert tdim2.shape == (4,)


class TestReshapedDimension(object):
    """Check functionality of class `ReshapedDimension`."""

    def test_transform(self, rdim):
        """Check method `transform`."""
        a = numpy.zeros((3, 2))
        a[0, 1] = 2
        assert rdim.transform([a, None]) == 2

    def test_reverse(self, rdim):
        """Check method `reverse`."""
        a = numpy.zeros((3, 2))
        a[0, 1] = 2
        p = a.reshape(-1).tolist() + [None]
        numpy.testing.assert_equal(rdim.reverse(p, 0), a)

    def test_interval(self, rdim):
        """Check method `interval`."""
        assert rdim.interval() == (
            -numpy.array(numpy.inf).astype(int) + 1,
            numpy.array(numpy.inf).astype(int) - 1,
        )

    def test_interval_from_categorical(self, rdim2):
        """Check how we should treat interval when original dimension is categorical."""
        assert rdim2.interval() == (0, 1)

    def test_eq(self, rdim, rdim2):
        """Return True if other is the same transformed dimension as self"""
        assert rdim != rdim2
        assert rdim == copy.deepcopy(rdim)

    def test_hash(self, rdim, rdim2):
        """Test that hash is consistent for identical and different transformed dimensions"""
        assert hash(rdim) != hash(rdim2)
        assert hash(rdim) == hash(copy.deepcopy(rdim))

    def test_get_hashable_members(self, rdim, rdim2):
        """Test that hashable members of the transformed dimensions are the aggregation of
        transformer's and original dimension's hashable members.
        """
        assert rdim._get_hashable_members() == (
            "View",
            "integer",
            "integer",
            "Compose",
            "Quantize",
            "real",
            "integer",
            "Identity",
            "real",
            "real",
            "yolo0",
            (3, 2),
            "real",
            (0.9,),
            (),
            None,
            "norm",
        )
        assert rdim2._get_hashable_members() == (
            "View",
            "real",
            "real",
            "Compose",
            "OneHotEncode",
            "integer",
            "real",
            4,
            "Compose",
            "Enumerate",
            "categorical",
            "integer",
            "Identity",
            "categorical",
            "categorical",
            "yolo2",
            (),
            "categorical",
            (),
            (),
            None,
            "Distribution",
        )

    def test_repr(self, rdim):
        """Check method `__repr__`."""
        assert (
            str(rdim)
            == "View(shape=(3, 2), index=(0, 1), Quantize(Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)))"
        )  # noqa

    def test_name_property(self, rdim):
        """Check property `name`."""
        assert rdim.name == "yolo0[0,1]"

    def test_type_property(self, rdim, rdim2):
        """Check property `type`."""
        assert rdim.type == "integer"
        assert rdim2.type == "real"

    def test_prior_name_property(self, rdim, rdim2):
        """Check property `prior_name`."""
        assert rdim.prior_name == "norm"
        assert rdim2.prior_name == "choices"

    def test_shape_property(self, rdim, rdim2):
        """Check property `shape`."""
        assert rdim.original_dimension.shape == (3, 2)
        assert rdim.shape == ()
        assert rdim2.original_dimension.shape == (4,)
        assert rdim2.shape == ()


@pytest.fixture(scope="module")
def space(dim, dim2, dim3):
    """Create an example `Space`."""
    space = Space()
    space.register(dim)
    space.register(dim2)
    space.register(dim3)
    return space


@pytest.fixture(scope="module")
def tspace(space, tdim, tdim2, tdim3):
    """Create an example `TransformedSpace`."""
    tspace = TransformedSpace(space)
    tspace.register(tdim)
    tspace.register(tdim2)
    tspace.register(tdim3)
    return tspace


@pytest.fixture(scope="module")
def rspace(tspace, rdims, rdims2, rdims3):
    """Create an example `ReshapedSpace`."""
    rspace = ReshapedSpace(tspace)
    for dim in itertools.chain(rdims.values(), rdims2.values(), rdims3.values()):
        rspace.register(dim)

    return rspace


class TestTransformedSpace(object):
    """Check functionality of class `TransformedSpace`."""

    def test_extends_space(self, tspace):
        """Check that `TransformedSpace` is actually a `Space`."""
        assert isinstance(tspace, Space)

    def test_transform(self, space, tspace, seed):
        """Check method `transform`."""
        yo = space.sample(seed=seed)[0]
        tyo = tspace.transform(yo)
        assert tyo in tspace

    def test_reverse(self, space, tspace, seed):
        """Check method `reverse`."""
        tyo = tspace.sample(seed=seed)[0]
        yo = tspace.reverse(tyo)
        assert yo in space

    def test_sample(self, space, tspace, seed):
        """Check method `sample`."""
        points = tspace.sample(n_samples=2, seed=seed)
        # pytest.set_trace()
        assert len(points) == 2
        assert points[0] in tspace
        assert points[1] in tspace
        assert tspace.reverse(points[0]) in space
        assert tspace.reverse(points[1]) in space


class TestReshapedSpace(object):
    """Check functionality of class `ReshapeSpace`."""

    def test_reverse(self, space, tspace, rspace, seed):
        """Check method `reverse`."""
        ryo = (
            numpy.zeros(tspace["yolo0"].shape).reshape(-1).tolist()
            + numpy.zeros(tspace["yolo2"].shape).reshape(-1).tolist()
            + [10]
        )
        yo = rspace.reverse(ryo)
        assert yo in space

    def test_contains(self, tspace, rspace, seed):
        """Check method `transform`."""
        ryo = (
            numpy.zeros(tspace["yolo0"].shape).reshape(-1).tolist()
            + numpy.zeros(tspace["yolo2"].shape).reshape(-1).tolist()
            + [10]
        )

        assert ryo in rspace

    def test_transform(self, space, rspace, seed):
        """Check method `transform`."""
        yo = space.sample(seed=seed)[0]
        tyo = rspace.transform(yo)
        assert tyo in rspace

    def test_sample(self, space, rspace, seed):
        """Check method `sample`."""
        points = rspace.sample(n_samples=2, seed=seed)
        assert len(points) == 2
        assert points[0] in rspace
        assert points[1] in rspace
        assert rspace.reverse(points[0]) in space
        assert rspace.reverse(points[1]) in space

    def test_interval(self, rspace):
        """Check method `interval`."""
        interval = rspace.interval()
        assert len(interval) == 3 * 2 + 4 + 1
        for i in range(3 * 2):
            # assert interval[i] == (-float('inf'), float('inf'))
            assert interval[i] == (
                -numpy.array(numpy.inf).astype(int) + 1,
                numpy.array(numpy.inf).astype(int) - 1,
            )
        for i in range(3 * 2, 3 * 2 + 4):
            assert interval[i] == (0, 1)
        assert interval[-1] == (3, 10)

    def test_reshape(self, rspace):
        """Verify that the dimension are reshaped properly, forward and backward"""
        point = [numpy.arange(6).reshape(3, 2), "3", 10]
        rpoint = point[0].reshape(-1).tolist() + [0.0, 0.0, 1.0, 0.0] + [10]
        assert rspace.transform(point) == tuple(rpoint)
        numpy.testing.assert_equal(rspace.reverse(rpoint)[0], point[0])
        assert rspace.reverse(rpoint)[1] == point[1]
        assert rspace.reverse(rpoint)[2] == point[2]

    def test_cardinality(self, dim2):
        """Check cardinality of reshaped space"""
        space = Space()
        space.register(Real("yolo0", "uniform", 0, 2, shape=(2, 2)))
        space.register(dim2)

        rspace = build_required_space(space, shape_requirement="flattened")
        assert rspace.cardinality == numpy.inf

        rspace = build_required_space(
            space, type_requirement="integer", shape_requirement="flattened"
        )
        assert rspace.cardinality == (3 ** (2 * 2)) * 4


@pytest.fixture(scope="module")
def space_each_type(dim, dim2, dim3, logdim, logintdim):
    """Create an example `Space`."""
    space = Space()
    space.register(dim)
    space.register(dim2)
    space.register(dim3)
    space.register(logdim)
    space.register(logintdim)
    return space


class TestRequiredSpaceBuilder(object):
    """Check functionality of builder function `build_required_space`."""

    @pytest.mark.xfail(
        reason="Bring it back when testing new builder and extend to shape and dist"
    )
    def test_not_supported_requirement(self, space_each_type):
        """Require something which is not supported."""
        with pytest.raises(TypeError) as exc:
            build_required_space(space_each_type, type_requirement="fasdfasf")
        assert "Unsupported" in str(exc.value)

    def test_no_requirement(self, space_each_type):
        """Check what is built using 'None' requirement."""
        tspace = build_required_space(space_each_type)
        assert len(tspace) == 5
        assert tspace[0].type == "real"
        assert tspace[1].type == "categorical"
        # NOTE:HEAD
        assert tspace[2].type == "integer"
        assert tspace[3].type == "real"
        assert tspace[4].type == "integer"
        assert (
            str(tspace)
            == """\
Space([Precision(4, Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),
       Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None),
       Integer(name=yolo3, prior={uniform: (3, 7), {}}, shape=(), default value=None),
       Precision(4, Real(name=yolo4, prior={reciprocal: (1.0, 10.0), {}}, shape=(3, 2), default value=None)),
       Integer(name=yolo5, prior={reciprocal: (1, 10), {}}, shape=(3, 2), default value=None)])\
"""
        )  # noqa

    def test_integer_requirement(self, space_each_type):
        """Check what is built using 'integer' requirement."""
        tspace = build_required_space(space_each_type, type_requirement="integer")
        assert len(tspace) == 5
        assert tspace[0].type == "integer"
        assert tspace[1].type == "integer"
        assert tspace[2].type == "integer"
        assert tspace[3].type == "integer"
        assert tspace[4].type == "integer"
        assert (
            str(tspace)
            == """\
Space([Quantize(Precision(4, Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None))),
       Enumerate(Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None)),
       Integer(name=yolo3, prior={uniform: (3, 7), {}}, shape=(), default value=None),
       Quantize(Precision(4, Real(name=yolo4, prior={reciprocal: (1.0, 10.0), {}}, shape=(3, 2), default value=None))),
       Integer(name=yolo5, prior={reciprocal: (1, 10), {}}, shape=(3, 2), default value=None)])\
"""
        )  # noqa

    def test_real_requirement(self, space_each_type):
        """Check what is built using 'real' requirement."""
        tspace = build_required_space(space_each_type, type_requirement="real")
        assert len(tspace) == 5
        assert tspace[0].type == "real"
        assert tspace[1].type == "real"
        assert tspace[2].type == "real"
        assert tspace[3].type == "real"
        assert tspace[4].type == "real"
        assert (
            str(tspace)
            == """\
Space([Precision(4, Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),
       OneHotEncode(Enumerate(Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None))),
       ReverseQuantize(Integer(name=yolo3, prior={uniform: (3, 7), {}}, shape=(), default value=None)),
       Precision(4, Real(name=yolo4, prior={reciprocal: (1.0, 10.0), {}}, shape=(3, 2), default value=None)),
       ReverseQuantize(Integer(name=yolo5, prior={reciprocal: (1, 10), {}}, shape=(3, 2), default value=None))])\
"""
        )  # noqa

    def test_numerical_requirement(self, space_each_type):
        """Check what is built using 'integer' requirement."""
        tspace = build_required_space(space_each_type, type_requirement="numerical")
        assert len(tspace) == 5
        assert tspace[0].type == "real"
        assert tspace[1].type == "integer"
        assert tspace[2].type == "integer"
        assert tspace[3].type == "real"
        assert tspace[4].type == "integer"
        assert (
            str(tspace)
            == """\
Space([Precision(4, Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),
       Enumerate(Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None)),
       Integer(name=yolo3, prior={uniform: (3, 7), {}}, shape=(), default value=None),
       Precision(4, Real(name=yolo4, prior={reciprocal: (1.0, 10.0), {}}, shape=(3, 2), default value=None)),
       Integer(name=yolo5, prior={reciprocal: (1, 10), {}}, shape=(3, 2), default value=None)])\
"""
        )  # noqa

    def test_linear_requirement(self, space_each_type):
        """Check what is built using 'linear' requirement."""
        tspace = build_required_space(space_each_type, dist_requirement="linear")
        assert len(tspace) == 5
        assert tspace[0].type == "real"
        assert tspace[1].type == "categorical"
        assert tspace[2].type == "integer"
        assert tspace[3].type == "real"
        assert tspace[4].type == "integer"
        assert (
            str(tspace)
            == """\
Space([Precision(4, Real(name=yolo0, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),
       Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None),
       Integer(name=yolo3, prior={uniform: (3, 7), {}}, shape=(), default value=None),
       Linearize(Precision(4, Real(name=yolo4, prior={reciprocal: (1.0, 10.0), {}}, shape=(3, 2), default value=None))),
       Quantize(Linearize(ReverseQuantize(Integer(name=yolo5, prior={reciprocal: (1, 10), {}}, shape=(3, 2), default value=None))))])\
"""
        )  # noqa

    def test_flatten_requirement(self, space_each_type):
        """Check what is built using 'flatten' requirement."""
        tspace = build_required_space(space_each_type, shape_requirement="flattened")

        # 1 integer + 1 categorical + 1 * (3, 2) shapes
        assert len(tspace) == 1 + 1 + 3 * (3 * 2)
        assert str(tspace).count("View") == 3 * (3 * 2)

        i = 0
        for _ in range(3 * 2):
            assert tspace[i].type == "real"
            i += 1

        assert tspace[i].type == "categorical"
        i += 1

        assert tspace[i].type == "integer"
        i += 1

        for _ in range(3 * 2):
            assert tspace[i].type == "real"
            i += 1

        for _ in range(3 * 2):
            assert tspace[i].type == "integer"
            i += 1

        tspace = build_required_space(
            space_each_type, shape_requirement="flattened", type_requirement="real"
        )

        # 1 integer + 4 categorical + 1 * (3, 2) shapes
        assert len(tspace) == 1 + 4 + 3 * (3 * 2)
        assert str(tspace).count("View") == 4 + 3 * (3 * 2)

    def test_capacity(self, space_each_type):
        """Check transformer space capacity"""
        tspace = build_required_space(space_each_type, type_requirement="real")
        assert tspace.cardinality == numpy.inf

        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ("asdfa", 2, 3, 4)
        dim = Categorical("yolo0", OrderedDict(zip(categories, probs)), shape=2)
        space.register(dim)
        dim = Integer("yolo2", "uniform", -3, 6)
        space.register(dim)
        tspace = build_required_space(space, type_requirement="integer")
        assert tspace.cardinality == (4 ** 2) * (6 + 1)

        dim = Integer("yolo3", "uniform", -3, 6, shape=(2, 1))
        space.register(dim)
        tspace = build_required_space(space, type_requirement="integer")
        assert tspace.cardinality == (4 ** 2) * (6 + 1) * ((6 + 1) ** (2 * 1))

        tspace = build_required_space(
            space, type_requirement="integer", shape_requirement="flattened"
        )
        assert tspace.cardinality == (4 ** 2) * (6 + 1) * ((6 + 1) ** (2 * 1))

        tspace = build_required_space(
            space, type_requirement="integer", dist_requirement="linear"
        )
        assert tspace.cardinality == (4 ** 2) * (6 + 1) * ((6 + 1) ** (2 * 1))


def test_quantization_does_not_violate_bounds():
    """Regress on bug that converts valid float in tdim to non valid excl. upper bound."""
    dim = Integer("yo", "uniform", 3, 7)
    transformers = [Reverse(Quantize())]
    tdim = TransformedDimension(Compose(transformers, dim.type), dim)
    assert 11 not in dim
    assert 10 in dim
    # but be careful, because upper bound is inclusive
    assert 11.5 not in tdim
    assert 10.6 in tdim
    assert tdim.reverse(9.6) in dim
    # solution is to quantize with 'floor' instead of 'round'
    assert tdim.reverse(9.6) == 9


def test_precision_with_linear(space, logdim, logintdim):
    """Test that precision isn't messed up by linearization."""
    space.register(logdim)
    space.register(logintdim)

    # Force precision on all real or linearized dimensions
    space["yolo0"].precision = 3
    space["yolo4"].precision = 4
    space["yolo5"].precision = 5

    # Create a point
    point = list(space.sample(1)[0])
    real_index = list(space.keys()).index("yolo0")
    logreal_index = list(space.keys()).index("yolo4")
    logint_index = list(space.keys()).index("yolo5")
    point[real_index] = 0.133333
    point[logreal_index] = 0.1222222
    point[logint_index] = 2

    # Check first without linearization
    tspace = build_required_space(space, type_requirement="numerical")
    # Check that transform is fine
    tpoint = tspace.transform(point)
    assert tpoint[real_index] == 0.133
    assert tpoint[logreal_index] == 0.1222
    assert tpoint[logint_index] == 2

    # Check that reserve does not break precision
    rpoint = tspace.reverse(tpoint)
    assert rpoint[real_index] == 0.133
    assert rpoint[logreal_index] == 0.1222
    assert rpoint[logint_index] == 2

    # Check with linearization
    tspace = build_required_space(
        space, dist_requirement="linear", type_requirement="real"
    )
    # Check that transform is fine
    tpoint = tspace.transform(point)
    assert tpoint[real_index] == 0.133
    assert tpoint[logreal_index] == numpy.log(0.1222)
    assert tpoint[logint_index] == numpy.log(2)

    # Check that reserve does not break precision
    rpoint = tspace.reverse(tpoint)
    assert rpoint[real_index] == 0.133
    assert rpoint[logreal_index] == 0.1222
    assert rpoint[logint_index] == 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.config`."""

import argparse
import logging
import os

import pytest
import yaml

from orion.core.io.config import Configuration, ConfigurationError


@pytest.fixture
def yaml_path():
    """Create a temporary yaml file and return the path"""
    file_path = "./my.yaml"
    with open(file_path, "w") as f:
        f.write(yaml.dump({"test": "from my yaml!"}))

    yield file_path

    os.remove(file_path)


@pytest.fixture
def broken_yaml_path():
    """Create a temporary yaml file and return the path"""
    file_path = "./my.yaml"
    with open(file_path, "w") as f:
        f.write(yaml.dump({"coucou": "from my yaml!"}))

    yield file_path

    os.remove(file_path)


@pytest.fixture
def subdict_yaml_path():
    """Create a temporary yaml file with subdicts and return the path"""
    file_path = "./my.yaml"
    with open(file_path, "w") as f:
        f.write(
            yaml.dump(
                {
                    "test": {"i am": "a sub-dict"},
                    "test2": {"test": {"me is": "sub-conf sub-dict"}},
                }
            )
        )

    yield file_path

    os.remove(file_path)


def test_fetch_non_existing_option():
    """Test that access to a non existing key raises ConfigurationError"""
    config = Configuration()
    with pytest.raises(ConfigurationError) as exc:
        config.voici_voila

    assert "Configuration does not have an attribute 'voici_voila'." in str(exc.value)


def test_access_to_config():
    """Test that access to _config returns properly _config.

    This is because getattr() could grasp any key including `_config` and makes
    it impossible to access the later
    """
    assert Configuration()._config == {}


def test_set_subconfig():
    """Test that setting a subconfig works"""
    config = Configuration()
    config.test = Configuration()

    assert isinstance(config.test, Configuration)

    with pytest.raises(ConfigurationError):
        config.test.voici_voila


def test_access_config_without_values():
    """Test that access to config without values raises ConfigurationError"""
    config = Configuration()
    config.add_option("test", option_type=int)
    with pytest.raises(ConfigurationError) as exc:
        config.test

    assert "Configuration not set and no default provided: test." in str(exc.value)


def test_set_non_existing_option():
    """Test that setting a non existing option crash"""
    config = Configuration()
    with pytest.raises(TypeError) as exc:
        config.test = 1
    assert "Can only set test as a Configuration, not <class 'int'>" in str(exc.value)


def test_set_subconfig_over_option():
    """Test that overwritting an option with a subconfig is not possible"""
    config = Configuration()
    config.add_option("test", option_type=int)
    config.test = 1
    assert config.test == 1
    with pytest.raises(TypeError) as exc:
        config.test = Configuration()
    assert "Cannot overwrite option test with a configuration" in str(exc.value)


def test_set_int_value():
    """Test that an integer option can have its value set"""
    config = Configuration()
    config.add_option("test", option_type=int)

    with pytest.raises(ConfigurationError) as exc:
        config.test

    config.test = 1
    assert config.test == 1
    config.test = "1"
    assert config.test == 1
    with pytest.raises(TypeError) as exc:
        config.test = "voici_voila"
    assert "<class 'int'> cannot be set to voici_voila with type <class 'str'>" in str(
        exc.value
    )


def test_set_real_value():
    """Test that a float option can have its value set"""
    config = Configuration()
    config.add_option("test", option_type=float)

    with pytest.raises(ConfigurationError) as exc:
        config.test

    config.test = 1
    assert config.test == 1.0
    config.test = "1"
    assert config.test == 1.0
    with pytest.raises(TypeError) as exc:
        config.test = "voici_voila"
    assert (
        "<class 'float'> cannot be set to voici_voila with type <class 'str'>"
        in str(exc.value)
    )


def test_set_str_value():
    """Test that a string option can have its value set"""
    config = Configuration()
    config.add_option("test", option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config.test = "1"
    assert config.test == "1"
    config.test = 1
    assert config.test == "1"


def test_set_value_of_subconfig_directly():
    """Test that we can access subconfig and set value directly"""
    config = Configuration()
    config.sub = Configuration()
    config.sub.add_option("test", option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config.sub.test = "1"
    assert config.sub.test == "1"
    config.sub.test = 1
    assert config.sub.test == "1"


def test_set_value_like_dict():
    """Test that we can set values like a dictionary"""
    config = Configuration()
    config.add_option("test", option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config["test"] = "1"
    assert config.test == "1"
    config["test"] = 1
    assert config.test == "1"


def test_set_subconfig_value_like_dict():
    """Test that we can set values like a dictionary"""
    config = Configuration()
    config.sub = Configuration()
    config.sub.add_option("test", option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config["sub.test"] = "1"
    assert config.sub.test == "1"
    config["sub.test"] = 1
    assert config.sub.test == "1"


def test_set_invalid_subconfig_value_like_dict():
    """Test that deep keys cannot be set if subconfig does not exist"""
    config = Configuration()
    with pytest.raises(BaseException) as exc:
        config["sub.test"] = "1"
    assert "Configuration does not have an attribute 'sub'." in str(exc.value)


def test_yaml_loading_empty_config(yaml_path):
    """Test loading for empty config fails like setting non existing attributes."""
    config = Configuration()

    with pytest.raises(ConfigurationError) as exc:
        config.load_yaml(yaml_path)

    assert "Configuration does not have an attribute 'test'." in str(exc.value)


def test_default_value():
    """Test that default value is given only when nothing else is available"""
    config = Configuration()
    config.add_option("test", option_type=str, default="voici_voila")
    assert config.test == "voici_voila"


def test_yaml_precedence(yaml_path):
    """Test that yaml definition has precedence over default values"""
    config = Configuration()
    config.add_option(
        "test", option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE"
    )
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"


def test_env_var_precedence(yaml_path):
    """Test that env_var has precedence over yaml values"""
    config = Configuration()
    config.add_option(
        "test", option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE"
    )
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"

    os.environ["TOP_SECRET_MESSAGE"] = "coussi_coussa"
    assert config.test == "coussi_coussa"
    del os.environ["TOP_SECRET_MESSAGE"]

    assert config.test == "from my yaml!"


def test_env_var_list(yaml_path):
    """Test that env_var lists are correctly handled"""
    config = Configuration()
    config.add_option(
        "test", option_type=list, default=["voici"], env_var="TOP_SECRET_LIST"
    )
    assert config.test == ["voici"]

    os.environ["TOP_SECRET_LIST"] = "voila:voici:voila"
    assert config.test == ["voila", "voici", "voila"]


def test_local_precedence(yaml_path):
    """Test local setting has precedence over env var values"""
    config = Configuration()
    config.add_option(
        "test", option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE"
    )
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"

    os.environ["TOP_SECRET_MESSAGE"] = "coussi_coussa"
    assert config.test == "coussi_coussa"

    config.test = "comme_ci_comme_ca"
    assert config.test == "comme_ci_comme_ca"

    del os.environ["TOP_SECRET_MESSAGE"]


def test_overwrite_subconfig():
    """Test that subconfig cannot be overwritten"""
    config = Configuration()
    config.nested = Configuration()
    with pytest.raises(ValueError) as exc:
        config.add_option("nested", option_type=str)
    assert "Configuration already contains nested" == str(exc.value)

    with pytest.raises(ValueError) as exc:
        config.nested = Configuration()
    assert "Configuration already contains subconfiguration nested" == str(exc.value)


def test_load_yaml_with_dict_items(subdict_yaml_path):
    """Test that yaml config with items assigned with dicts is supported"""
    config = Configuration()
    default = {"default": "sub-dict"}
    config.add_option("test", option_type=dict, default=default)
    assert config.test == default
    config.test2 = Configuration()
    config.test2.add_option("test", option_type=dict, default=default)
    assert config.test2.test == default

    config.load_yaml(subdict_yaml_path)
    assert config.test == {"i am": "a sub-dict"}
    assert config.test2.test == {"me is": "sub-conf sub-dict"}


def test_load_yaml_unknown_option(broken_yaml_path):
    """Test error message when yaml config contains unknown options"""
    config = Configuration()
    config.add_option("test", option_type=str, default="hello")
    assert config.test == "hello"

    with pytest.raises(ConfigurationError) as exc:
        config.load_yaml(broken_yaml_path)
    assert exc.match("Configuration does not have an attribute 'coucou'")


def test_to_dict():
    """Test dictionary representation of the configuration"""
    config = Configuration()
    config.add_option("test", option_type=str, default="voici_voila")
    config.nested = Configuration()
    config.nested.add_option("test2", option_type=str, default="zici")

    assert config.to_dict() == {"test": "voici_voila", "nested": {"test2": "zici"}}

    config.test = "hello"
    config.nested.test2 = "labas"

    assert config.to_dict() == {"test": "hello", "nested": {"test2": "labas"}}


def test_key_curation():
    """Test that both - and _ maps to same options"""
    config = Configuration()
    config.add_option("test-with-dashes", option_type=int, default=1)
    config.add_option("test_with_underscores", option_type=int, default=2)
    config.add_option("test-all_mixedup", option_type=int, default=3)

    assert config["test-with-dashes"] == 1
    assert config["test_with_underscores"] == 2
    assert config["test-all_mixedup"] == 3

    assert config["test_with_dashes"] == 1
    assert config["test-with-underscores"] == 2
    assert config["test_all-mixedup"] == 3

    assert config.test_with_dashes == 1
    assert config.test_with_underscores == 2
    assert config.test_all_mixedup == 3

    config["test_with_dashes"] = 4
    assert config["test-with-dashes"] == 4


def test_nested_key_curation():
    """Test that both - and _ maps to same options in nested configs as well"""
    config = Configuration()
    config.add_option("test-with-dashes", option_type=str, default="voici_voila")
    config.nested = Configuration()
    config.nested.add_option("test_with_underscores", option_type=str, default="zici")

    assert config["nested"]["test_with_underscores"] == "zici"
    assert config["nested"]["test-with-underscores"] == "zici"

    config["nested"]["test-with-underscores"] = "labas"
    assert config.nested.test_with_underscores == "labas"


def test_help_option():
    """Verify adding documentation to options."""
    config = Configuration()
    config.add_option("option", option_type=str, help="A useless option!")

    assert config.help("option") == "A useless option!"


def test_help_nested_option():
    """Verify adding documentation to a nested option."""
    config = Configuration()
    config.add_option("option", option_type=str, help="A useless option!")
    config.nested = Configuration()
    config.nested.add_option("option", option_type=str, help="A useless nested option!")

    assert config.help("nested.option") == "A useless nested option!"
    assert config.nested.help("option") == "A useless nested option!"


def test_help_option_with_default():
    """Verify adding documentation to options with default value."""
    config = Configuration()
    config.add_option("option", option_type=str, default="a", help="A useless option!")

    assert config.help("option") == "A useless option! (default: a)"


def test_no_help_option():
    """Verify not adding documentation to options."""
    config = Configuration()
    config.add_option("option", option_type=str)

    assert config.help("option") == "Undocumented"


def test_argument_parser():
    """Verify the argument parser built based on config."""
    config = Configuration()
    config.add_option("option", option_type=str)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args(["--option", "a"])

    assert options.option == "a"


def test_argument_parser_ignore_default():
    """Verify the argument parser does not get default values."""
    config = Configuration()
    config.add_option("option", option_type=str, default="b")

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args([])

    assert options.option is None


def test_argument_parser_rename():
    """Verify the argument parser built based on config with some options renamed."""
    config = Configuration()
    config.add_option("option", option_type=str)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser, rename=dict(option="--new-option"))

    with pytest.raises(SystemExit) as exc:
        options = parser.parse_args(["--option", "a"])

    assert exc.match("2")

    options = parser.parse_args(["--new-option", "a"])

    assert options.new_option == "a"


def test_argument_parser_dict_list_tuple():
    """Verify the argument parser does not contain options of type dict/list/tuple in config."""
    config = Configuration()
    config.add_option("st", option_type=str)
    config.add_option("di", option_type=dict)
    config.add_option("li", option_type=list)
    config.add_option("tu", option_type=tuple)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args([])
    assert vars(options) == {"st": None}

    with pytest.raises(SystemExit) as exc:
        options = parser.parse_args(["--di", "ct"])

    assert exc.match("2")


def test_deprecate_option(caplog):
    """Test deprecating an option."""
    config = Configuration()
    config.add_option(
        "option",
        option_type=str,
        default="hello",
        deprecate=dict(version="v1.0", alternative="None! T_T"),
    )

    config.add_option("ok", option_type=str, default="hillo")

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == "hello"

    assert caplog.record_tuples == [
        (
            "orion.core.io.config",
            logging.WARNING,
            "(DEPRECATED) Option `option` will be removed in v1.0. Use `None! T_T` instead.",
        )
    ]

    caplog.clear()

    # Access the non-deprecated option and trigger no warnings.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.ok == "hillo"

    assert caplog.record_tuples == []


def test_deprecate_option_missing_version():
    """Verify option deprecation if version is missing."""
    config = Configuration()
    with pytest.raises(ValueError) as exc:
        config.add_option(
            "option", option_type=str, deprecate=dict(alternative="None! T_T")
        )

    assert exc.match("`version` is missing in deprecate option")


def test_deprecate_option_no_alternative(caplog):
    """Verify option deprecation when there is no alternative."""
    config = Configuration()
    config.add_option(
        "option", option_type=str, default="hello", deprecate=dict(version="v1.0")
    )

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == "hello"

    assert caplog.record_tuples == [
        (
            "orion.core.io.config",
            logging.WARNING,
            "(DEPRECATED) Option `option` will be removed in v1.0.",
        )
    ]


def test_deprecate_option_help():
    """Verify help message of a deprecated option."""
    config = Configuration()
    config.add_option(
        "option",
        option_type=str,
        deprecate=dict(version="v1.0", alternative="None! T_T"),
        help="A useless option!",
    )

    assert config.help("option") == "(DEPRECATED) A useless option!"


def test_deprecate_option_print_with_different_name(caplog):
    """Verify deprecation warning with different name (for nested options)."""
    config = Configuration()
    config.add_option(
        "option",
        option_type=str,
        default="hello",
        deprecate=dict(version="v1.0", alternative="None! T_T", name="nested.option"),
    )

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == "hello"

    assert caplog.record_tuples == [
        (
            "orion.core.io.config",
            logging.WARNING,
            "(DEPRECATED) Option `nested.option` will be removed in v1.0. Use `None! T_T` instead.",
        )
    ]


def test_get_deprecated_key(caplog):
    """Verify deprecation warning using get()."""
    config = Configuration()
    config.add_option(
        "option",
        option_type=str,
        default="hello",
        deprecate=dict(version="v1.0", alternative="None! T_T"),
    )

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.get("option") == "hello"

    assert caplog.record_tuples == [
        (
            "orion.core.io.config",
            logging.WARNING,
            "(DEPRECATED) Option `option` will be removed in v1.0. Use `None! T_T` instead.",
        )
    ]


def test_get_deprecated_key_ignore_warning(caplog):
    """Verify deprecation warning using get(deprecated='ignore')."""
    config = Configuration()
    config.add_option(
        "option",
        option_type=str,
        default="hello",
        deprecate=dict(version="v1.0", alternative="None! T_T"),
    )

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.get("option", deprecated="ignore") == "hello"

    assert caplog.record_tuples == []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.pickleddb`."""
import pytest
from test_database import clean_db, orion_db

from orion.core.io.database import Database
from orion.core.io.database.ephemeraldb import (
    EphemeralCollection,
    EphemeralDB,
    EphemeralDocument,
)


@pytest.fixture(scope="module", autouse=True)
def db_type(pytestconfig, request):
    """Return the string identifier of a EphemeralDB if the --mongodb option is
    not active"""
    if pytestconfig.getoption("--mongodb"):
        pytest.skip("ephemeraldb tests disabled")
    yield "ephemeraldb"


@pytest.fixture()
def document(db_type):
    """Return EphemeralDocument."""
    if db_type != "ephemeraldb":
        pytest.skip("ephemeraldb test only")
    yield EphemeralDocument({"_id": 1, "hello": "there", "mighty": "duck"})


@pytest.fixture()
def subdocument(db_type):
    """Return EphemeralDocument with a subdocument."""
    if db_type != "ephemeraldb":
        pytest.skip("ephemeraldb test only")
    yield EphemeralDocument(
        {"_id": 1, "hello": "there", "mighty": "duck", "and": {"the": "drake"}}
    )


@pytest.fixture()
def collection(document, db_type):
    """Return EphemeralCollection."""
    if db_type != "ephemeraldb":
        pytest.skip("ephemeraldb test only")

    collection = EphemeralCollection()
    collection.insert_many([document.to_dict()])

    yield collection


# TESTS SET


@pytest.mark.usefixtures("clean_db")
class TestIndex(object):
    """Test index for :meth:`orion.core.io.database.ephemeraldb.EphemeralCollection`."""

    def test_create_index(self, collection):
        """Test if new index added property."""
        collection.create_index("hello")
        assert collection._indexes == {"_id_": (("_id",), {(1,)})}

        collection.create_index("hello", unique=True)
        assert collection._indexes == {
            "_id_": (("_id",), {(1,)}),
            "hello_1": (("hello",), {("there",)}),
        }

    def test_track_index(self, collection):
        """Test if index values are tracked property."""
        collection.create_index("hello", unique=True)
        collection.insert_many([{"hello": "here"}, {"hello": 2}])
        assert collection._indexes == {
            "_id_": (("_id",), {(1,), (2,), (3,)}),
            "hello_1": (("hello",), {("there",), ("here",), (2,)}),
        }

    def test_index_over_non_existing_field(self, collection):
        """Test if index values are tracked property."""
        collection.create_index(
            [("hello", Database.DESCENDING), ("idontexist", Database.ASCENDING)],
            unique=True,
        )

        collection.insert_many([{"hello": "here"}, {"hello": 2}])
        assert collection._indexes == {
            "_id_": (("_id",), {(1,), (2,), (3,)}),
            "hello_1_idontexist_1": (
                ("hello", "idontexist"),
                {("there", None), ("here", None), (2, None)},
            ),
        }
        assert collection.find({}, selection={"hello": 1, "idontexist": 1}) == [
            {"_id": 1, "hello": "there", "idontexist": None},
            {"_id": 2, "hello": "here", "idontexist": None},
            {"_id": 3, "hello": 2, "idontexist": None},
        ]


@pytest.mark.usefixtures("clean_db")
class TestSelect(object):
    """Calls :meth:`orion.core.io.database.ephemeraldb.EphemeralDocument.select`."""

    def test_select_all(self, document):
        """Select only one field."""
        assert document.select({}) == {"_id": 1, "hello": "there", "mighty": "duck"}

    def test_select_id(self, document):
        """Select only one field."""
        assert document.select({"_id": 1}) == {"_id": 1}

    def test_select_one(self, document):
        """Select only one field."""
        assert document.select({"hello": 1}) == {"_id": 1, "hello": "there"}

    def test_select_two(self, document):
        """Select only two field."""
        assert document.select({"hello": 1, "mighty": 1}) == {
            "_id": 1,
            "hello": "there",
            "mighty": "duck",
        }

    def test_unselect_one(self, document):
        """Unselect only one field."""
        assert document.select({"hello": 0}) == {"_id": 1, "mighty": "duck"}

    def test_unselect_two(self, document):
        """Unselect two field."""
        assert document.select({"_id": 0, "hello": 0}) == {"mighty": "duck"}

    def test_mixed_select(self, document):
        """Select one field and unselect _id."""
        assert document.select({"_id": 0, "hello": 1}) == {"hello": "there"}

    def test_select_unexisting_field(self, document):
        """Select field that does not exist and should return None."""
        assert document.select({"idontexist": 1}) == {"_id": 1, "idontexist": None}


@pytest.mark.usefixtures("clean_db")
class TestMatch:
    """Calls :meth:`orion.core.io.database.ephemeraldb.EphemeralDocument.match`."""

    def test_match_eq(self, document):
        """Test eq operator"""
        assert document.match({"hello": "there"})
        assert not document.match({"hello": "not there"})

    def test_match_sub_eq(self, subdocument):
        """Test eq operator with sub document"""
        assert subdocument.match({"and.the": "drake"})
        assert not subdocument.match({"and.no": "drake"})

    def test_match_in(self, subdocument):
        """Test $in operator with document"""
        assert subdocument.match({"hello": {"$in": ["there", "here"]}})
        assert not subdocument.match({"hello": {"$in": ["ici", "here"]}})

    def test_match_sub_in(self, subdocument):
        """Test $in operator with sub document"""
        assert subdocument.match({"and.the": {"$in": ["duck", "drake"]}})
        assert not subdocument.match({"and.the": {"$in": ["hyppo", "lion"]}})

    def test_match_gte(self, document):
        """Test $gte operator with document"""
        assert document.match({"_id": {"$gte": 1}})
        assert document.match({"_id": {"$gte": 0}})
        assert not document.match({"_id": {"$gte": 2}})

    def test_match_gt(self, document):
        """Test $gt operator with document"""
        assert document.match({"_id": {"$gt": 0}})
        assert not document.match({"_id": {"$gt": 1}})

    def test_match_lte(self, document):
        """Test $lte operator with document"""
        assert document.match({"_id": {"$lte": 2}})
        assert document.match({"_id": {"$lte": 1}})
        assert not document.match({"_id": {"$lte": 0}})

    def test_match_ne(self, document):
        """Test $ne operator with document"""
        assert document.match({"hello": {"$ne": "here"}})
        assert not document.match({"hello": {"$ne": "there"}})

    def test_match_bad_operator(self, document):
        """Test invalid operator handling"""
        with pytest.raises(ValueError) as exc:
            document.match({"_id": {"$voici_voila": 0}})

        assert "Operator '$voici_voila' is not supported" in str(exc.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
WSGI REST server application
============================

Exposes a WSGI REST server application instance by subclassing ``falcon.API``.

"""

import falcon

from orion.serving.experiments_resource import ExperimentsResource
from orion.serving.plots_resources import PlotsResource
from orion.serving.runtime import RuntimeResource
from orion.serving.trials_resource import TrialsResource
from orion.storage.base import setup_storage


class WebApi(falcon.API):
    """
    Main entry point into a Falcon-based app. An instance provides a callable WSGI interface and a
    routing engine.
    """

    def __init__(self, config=None):
        super(WebApi, self).__init__()
        self.config = config

        setup_storage(config.get("storage"))

        # Create our resources
        root_resource = RuntimeResource()
        experiments_resource = ExperimentsResource()
        trials_resource = TrialsResource()
        plots_resource = PlotsResource()

        # Build routes
        self.add_route("/", root_resource)
        self.add_route("/experiments", experiments_resource)
        self.add_route("/experiments/{name}", experiments_resource, suffix="experiment")
        self.add_route(
            "/trials/{experiment_name}", trials_resource, suffix="trials_in_experiment"
        )
        self.add_route(
            "/trials/{experiment_name}/{trial_id}",
            trials_resource,
            suffix="trial_in_experiment",
        )
        self.add_route("/plots/lpi/{experiment_name}", plots_resource, suffix="lpi")
        self.add_route(
            "/plots/partial_dependencies/{experiment_name}",
            plots_resource,
            suffix="partial_dependencies",
        )
        self.add_route(
            "/plots/parallel_coordinates/{experiment_name}",
            plots_resource,
            suffix="parallel_coordinates",
        )
        self.add_route(
            "/plots/regret/{experiment_name}", plots_resource, suffix="regret"
        )

    def start(self):
        """A hook to when a Gunicorn worker calls run()."""
        pass

    def stop(self, signal):
        """A hook to when a Gunicorn worker starts shutting down."""
        pass



"""
===========
Parallelism
===========


Multiprocessing
---------------

(joblib's loky)


Dask
----


"""

import joblib
import numpy
from sklearn import datasets
from sklearn.model_selection import cross_validate
from sklearn.svm import SVC

from orion.client import create_experiment


def main(C, gamma, tol, class_weight):

    diabetes = datasets.load_diabetes()

    X = diabetes.data
    y = diabetes.target

    model = SVC(kernel="rbf", C=C, gamma=gamma, tol=tol, class_weight=class_weight)

    # Single metric evaluation using cross_validate
    with joblib.parallel_backend("dask"):
        cv_results = cross_validate(model, X, y, cv=5)

    accuracy = numpy.mean(cv_results["test_score"])
    error_rate = 1 - accuracy

    return [{"name": "test_error_rate", "type": "objective", "value": error_rate}]


def hpo(n_workers=16):

    experiment = create_experiment(
        name="dask",
        max_trials=1000,
        max_broken=5,
        space={
            "C": "loguniform(1e-6, 1e6, precision=None)",
            "gamma": "loguniform(1e-8, 1e8, precision=None)",
            "tol": "loguniform(1e-4, 1e-1, precision=None)",
            "class_weight": "choices([None, 'balanced'])",
        },
        algorithms={"random": {"seed": 1}},
    )

    with experiment.tmp_executor("dask", n_workers=n_workers):
        experiment.workon(main, n_workers=n_workers // 2)

    experiment.plot.regret().show()
    experiment.plot.partial_dependencies(params=["C", "gamma", "tol"]).show()


if __name__ == "__main__":
    hpo()
